#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The concrete model wrappers for parsing all TF model formats."""

import copy
import datetime
import importlib
import json
import os
import shutil
import sys
import tempfile
import time
from abc import abstractmethod

import numpy as np
import tensorflow as tf

from neural_compressor.common import logger
from neural_compressor.common.utils import DEFAULT_WORKSPACE
from neural_compressor.tensorflow.utils.utility import version1_gte_version2, version1_lt_version2

tensor_to_node = lambda s: list(set([x.split(":")[0] for x in s]))


def get_tf_model_type(model):
    """The interface of getting type of tensorflow models."""
    try:
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        model_type = get_model_type(model)
    except:  # pragma: no cover
        os.environ.pop("CUDA_DEVICE_ORDER")
        os.environ.pop("CUDA_VISIBLE_DEVICES")
        raise TypeError(
            "Tensorflow model format is not correctly detected. This could be"
            + "caused by unsupported model or inappropriate framework installation."
        )
    else:
        return model_type


def get_model_type(model):
    """Get Tensorflow mode type.

    Args:
        model (string or model object): model path or model object.

    Returns:
        string: model type
    """
    from neural_compressor.tensorflow.quantization.utils.utility import is_ckpt_format, is_saved_model_format

    if isinstance(model, str):
        model = os.path.abspath(os.path.expanduser(model))
        if (
            ((model.endswith(".h5") or model.endswith(".keras")) and os.path.isfile(model))
            or is_saved_model_format(os.path.dirname(model))
            or (os.path.isdir(model) and is_saved_model_format(model))
        ):
            if version1_lt_version2(tf.version.VERSION, "2.10.0"):  # pragma: no cover
                logger.warning("keras model running on tensorflow 2.10.0 and" " lower not support intel ITEX.")
            try:
                model = tf.keras.models.load_model(model)
                if isinstance(model, tf.keras.Model) and hasattr(model, "to_json"):
                    return "keras"
                return "saved_model"  # pragma: no cover
            except:
                pass
    if isinstance(model, tf.keras.Model) and hasattr(model, "to_json"):
        if json.loads(model.to_json())["class_name"] in ["Sequential", "Functional"]:
            # Keras adaptor only support Sequential or Functional model
            return "keras"
        else:  # pragma: no cover
            # otherwise, the backend will fallback to tensorflow_itex
            return "saved_model"
    if isinstance(model, tf.Graph):
        return "graph"
    elif isinstance(model, tf.compat.v1.GraphDef):
        return "graph_def"
    elif not version1_gte_version2(tf.version.VERSION, "2.16.1") and isinstance(
        model, tf.compat.v1.estimator.Estimator
    ):  # pragma: no cover
        return "estimator"
    elif isinstance(model, str):
        model = os.path.abspath(os.path.expanduser(model))
        if model.endswith(".pb") and os.path.isfile(model):
            if is_saved_model_format(os.path.dirname(model)):  # pragma: no cover
                return "saved_model"
            else:
                return "frozen_pb"
        elif model.endswith(".ckpt") and os.path.isfile(model):  # pragma: no cover
            return "slim"
        elif os.path.isdir(model):
            if is_ckpt_format(model):
                return "checkpoint"
            elif is_saved_model_format(model):
                return "saved_model"
        elif os.path.isfile(model + ".pb"):
            return "frozen_pb"

    raise ValueError("model {} has not recognized model type....".format(model))


def validate_graph_node(graph_def, node_names):
    """Validate nodes exist in the graph_def.

    Args:
        graph_def (tf.compat.v1.GraphDef): tf.compat.v1.GraphDef object.
        node_names (list of string): node names to be validated.
    """
    if len(node_names) == 0:
        return False
    all_node_name = [node.name for node in graph_def.node]
    for user_name in node_names:
        if user_name not in all_node_name:
            logger.warning(str("Node name {} specified in yaml doesn't exist in the model.").format(user_name))
            return False
    return True


def validate_and_inference_input_output(graph_def, input_tensor_names, output_tensor_names):
    """Validate and inference the input and output tensor names of graph_def.

    Args:
        graph_def (tf.compat.v1.GraphDef): tf.compat.v1.GraphDef object.
        input_tensor_names (list of string): input_tensor_names of graph_def.
        output_tensor_names (list of string): output_tensor_names of graph_def.

    Returns:
        input_tensor_names (list of string): validated input_tensor_names.
        output_tensor_names (list of string): validated output_tensor_names.
    """
    from neural_compressor.tensorflow.quantization.utils.utility import get_input_output_node_names

    temp_output_tensor_names = []
    if validate_graph_node(graph_def, tensor_to_node(input_tensor_names)):
        input_tensor_names = input_tensor_names
    else:
        input_tensor_names, temp_output_tensor_names = get_input_output_node_names(graph_def)

    if validate_graph_node(graph_def, tensor_to_node(output_tensor_names)):
        output_tensor_names = output_tensor_names
    elif temp_output_tensor_names:
        output_tensor_names = temp_output_tensor_names
    else:  # pragma: no cover
        _, output_tensor_names = get_input_output_node_names(graph_def)

    return input_tensor_names, output_tensor_names


def graph_session(model, input_tensor_names, output_tensor_names, **kwargs):
    """Helper to build session with tf.compat.v1.Graph.

    Args:
        model (tf.compat.v1.Graph): tf.compat.v1.Graph object.
        input_tensor_names (list of string): input_tensor_names of model.
        output_tensor_names (list of string): output_tensor_names of model.

    Returns:
        sess (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_tensor_names (list of string): validated input_tensor_names.
        output_tensor_names (list of string): validated output_tensor_names.
    """
    config = tf.compat.v1.ConfigProto()
    config.use_per_session_threads = 1
    config.inter_op_parallelism_threads = 1
    sess = tf.compat.v1.Session(graph=model, config=config)

    input_tensor_names, output_tensor_names = validate_and_inference_input_output(
        model.as_graph_def(), input_tensor_names, output_tensor_names
    )

    return sess, input_tensor_names, output_tensor_names


def graph_def_session(model, input_tensor_names, output_tensor_names, **kwargs):
    """Build session with tf.compat.v1.GraphDef.

    Args:
        model (tf.compat.v1.GraphDef): tf.compat.v1.GraphDef object.
        input_tensor_names (list of string): input_tensor_names of model.
        output_tensor_names (list of string): output_tensor_names of model.

    Returns:
        sess (tf.compat.v1.Session): tf.compat.v1.Session object
        input_tensor_names (list of string): validated input_tensor_names
        output_tensor_names (list of string): validated output_tensor_names
    """
    device = kwargs.get("device")
    graph = tf.Graph()
    if version1_lt_version2(tf.version.VERSION, "2.0.0"):  # pragma: no cover
        from tensorflow._api.v1.config import experimental

        list_physical_devices = experimental.list_physical_devices
    else:
        list_physical_devices = tf.config.list_physical_devices

    try:
        with graph.as_default():
            tf.import_graph_def(model, name="")
    except:
        input_tensor_names, output_tensor_names = validate_and_inference_input_output(
            model, input_tensor_names, output_tensor_names
        )
        from neural_compressor.tensorflow.quantization.utils.utility import (
            fix_ref_type_of_graph_def,
            strip_unused_nodes,
        )

        model = fix_ref_type_of_graph_def(model)
        input_node_names = tensor_to_node(input_tensor_names)
        output_node_names = tensor_to_node(output_tensor_names)
        model = strip_unused_nodes(model, input_node_names, output_node_names)
        with graph.as_default():
            tf.import_graph_def(model, name="")

    return graph_session(graph, input_tensor_names, output_tensor_names, **kwargs)


def frozen_pb_session(model, input_tensor_names, output_tensor_names, **kwargs):
    """Build session with frozen pb.

    Args:
        model (string): model path.
        input_tensor_names (list of string): input_tensor_names of model.
        output_tensor_names (list of string): output_tensor_names of model.

    Returns:
        sess (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_tensor_names (list of string): validated input_tensor_names.
        output_tensor_names (list of string): validated output_tensor_names.
    """
    graph_def = tf.compat.v1.GraphDef()
    model = model if model.endswith(".pb") else model + ".pb"
    with open(model, "rb") as f:
        graph_def.ParseFromString(f.read())
    return graph_def_session(graph_def, input_tensor_names, output_tensor_names, **kwargs)


def _contains_function_with_implements_attr(saved_model_proto):
    meta_graph = saved_model_proto.meta_graphs[0]
    for function in meta_graph.graph_def.library.function:
        if function.attr.get("_implements", None) or function.attr.get("api_implements", None):  # pragma: no cover
            return True
    return False


def load_saved_model(model, saved_model_tags, input_tensor_names, output_tensor_names):  # pragma: no cover
    """Load graph_def from saved model with the default serving signature key.

    Args:
        model: Directory of the SavedModel.
        saved_model_tags: Set of tags identifying the MetaGraphDef within the
            SavedModel to analyze.
        input_tensor_names (list of string): input_tensor_names of model.
        output_tensor_names (list of string): output_tensor_names of model.

    Returns:
        graph_def: The loaded GraphDef.
        input_tensors: List of input tensors.
        output_tensors: List of output tensors.
    """
    config = tf.compat.v1.ConfigProto()
    config.use_per_session_threads = 1
    config.inter_op_parallelism_threads = 1
    if not os.listdir(os.path.join(model, "variables")):
        sess = tf.compat.v1.Session(graph=tf.Graph(), config=config)
        loader = tf.compat.v1.saved_model.loader.load(sess, ["serve"], model)
        if len(input_tensor_names) == 0:
            input_tensor_names = [i.name for _, i in loader.signature_def["serving_default"].inputs.items()]
        else:
            assert validate_graph_node(
                sess.graph.as_graph_def(), tensor_to_node(input_tensor_names)
            ), "tensor names {} not in the graph".format(input_tensor_names)

        if len(output_tensor_names) == 0:
            output_tensor_names = [i.name for _, i in loader.signature_def["serving_default"].outputs.items()]
        else:
            assert validate_graph_node(
                sess.graph.as_graph_def(), tensor_to_node(output_tensor_names)
            ), "tensor names {} not in the graph".format(output_tensor_names)

        return sess.graph.as_graph_def(), input_tensor_names, output_tensor_names
    else:
        from tensorflow.core.protobuf import config_pb2, meta_graph_pb2
        from tensorflow.python.eager import context
        from tensorflow.python.framework.convert_to_constants import convert_variables_to_constants_v2
        from tensorflow.python.grappler import tf_optimizer
        from tensorflow.python.saved_model import load, signature_constants, tag_constants
        from tensorflow.python.training import saver

        _saved_model = load.load(model, [tag_constants.SERVING])
        func = _saved_model.signatures[signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY]
        frozen_func = convert_variables_to_constants_v2(func)
        grappler_meta_graph_def = saver.export_meta_graph(
            graph_def=frozen_func.graph.as_graph_def(), graph=frozen_func.graph
        )
        if len(input_tensor_names) == 0:
            input_tensor_names = [i.name.split(":")[0] for i in frozen_func.inputs]
        if len(output_tensor_names) == 0:
            output_tensor_names = [i.name.split(":")[0] for i in frozen_func.outputs]
        # Add a collection 'train_op' so that Grappler knows the outputs.
        fetch_collection = meta_graph_pb2.CollectionDef()
        for array in frozen_func.inputs + frozen_func.outputs:
            fetch_collection.node_list.value.append(array.name)
            grappler_meta_graph_def.collection_def["train_op"].CopyFrom(fetch_collection)
        grappler_session_config = config_pb2.ConfigProto()
        rewrite_options = grappler_session_config.graph_options.rewrite_options
        rewrite_options.min_graph_nodes = -1
        opt = tf_optimizer.OptimizeGraph(grappler_session_config, grappler_meta_graph_def, graph_id=b"tf_graph")
        return opt, input_tensor_names, output_tensor_names


def _get_graph_from_saved_model_v3(model, input_tensor_names, output_tensor_names):
    """The version 3 function that get graph from saved_model.

    Args:
        model (string or tf.keras.Model): model path or tf.keras.Model object.
        input_tensor_names (list of string): input tensor names of the model.
        output_tensor_names (list of string): output tensor names of the model.

    Returns:
        graph_def (tf.compat.v1.Session): tf.compat.v1.Session object.
        inputs (list of string): validated input names.
        outputs (list of string): validated output names.
    """
    from neural_compressor.adaptor.tf_utils.util import parse_saved_model

    if isinstance(model, tf.keras.Model):  # pragma: no cover
        save_folder = "/tmp_model.keras" if version1_gte_version2(tf.version.VERSION, "2.16.1") else "/saved_model"
        tmp_dir = DEFAULT_WORKSPACE + save_folder
        model.save(tmp_dir)
        model = tmp_dir
    graph_def, _, _, _, input_names, output_names = parse_saved_model(
        model, True, input_tensor_names, output_tensor_names
    )

    return graph_def, input_names, output_names


def _get_graph_from_saved_model_v2(saved_model_dir, input_tensor_names, output_tensor_names):
    """The version 2 function that get graph from the original keras model.

    Args:
        saved_model_dir (string): model path of a temporary saved_model.
        input_tensor_names (list of string): input tensor names of the model.
        output_tensor_names (list of string): output tensor names of the model.

    Returns:
        graph_def (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_names (list of string): validated input names.
        output_names (list of string): validated output names.
    """
    from tensorflow.python.saved_model import signature_constants, tag_constants

    saved_model_exported_names = [signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY]
    saved_model_tags = set([tag_constants.SERVING])

    return load_saved_model(saved_model_dir, saved_model_tags, input_tensor_names, output_tensor_names)


def _get_graph_from_original_keras_v2(model):  # pragma: no cover
    """The version 2 function that get graph from the original keras model.

    Args:
        model (string or tf.keras.Model): model path or tf.keras.Model object.

    Returns:
        graph_def (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_names (list of string): validated input names.
        output_names (list of string): validated output names.
    """
    from tensorflow.lite.python.convert import OpsSet
    from tensorflow.lite.python.util import (
        get_grappler_config,
        model_input_signature,
        run_graph_optimizations,
        trace_model_call,
    )
    from tensorflow.python.eager import def_function
    from tensorflow.python.framework import convert_to_constants, dtypes

    input_signature = None
    # If the model's call is not a `tf.function`, then we need to first get its
    # input signature from `model_input_signature` method.
    if not isinstance(model.call, def_function.Function):
        input_signature = model_input_signature(model, keep_original_batch_size=False)

    func = trace_model_call(model, input_signature)
    concrete_func = func.get_concrete_function()
    funcs = [concrete_func]

    frozen_func, graph_def = convert_to_constants.convert_variables_to_constants_v2_as_graph(
        funcs[0], lower_control_flow=False
    )

    input_tensors = [tensor for tensor in frozen_func.inputs if tensor.dtype != dtypes.resource]
    output_tensors = frozen_func.outputs
    # Grappler will also try to lower while loop into switch merge
    # representation which is undesired for Ophints, so we simply remove
    # those attributes to prevent Grappler from doing so.
    graph = convert_to_constants.disable_lower_using_switch_merge(graph_def)
    # Run function inlining optimization to ensure any models generated
    # through the from_frozen_graph path have been inlined.
    # grappler_config = get_grappler_config(['function'])
    # graph_def = run_graph_optimizations(
    #             graph,
    #             input_tensors,
    #             output_tensors,
    #             config=grappler_config)
    input_names = [tensor.name.split(":")[0] for tensor in input_tensors]
    output_names = [tensor.name.split(":")[0] for tensor in output_tensors]
    return graph_def, input_names, output_names


def _check_keras_format(model, saved_model_dir):  # pragma: no cover
    """Decide which method will be used to get graph from the saved_model .

    Args:
        model (string or tf.keras.Model): model path or tf.keras.Model object.
        saved_model_dir (string): the path to save a temporary saved_model.

    Returns:
        graph_def (tf.compat.v1.Session): tf.compat.v1.Session object.
        inputs (list of string): validated input names.
        outputs (list of string): validated output names.
    """
    from tensorflow.python import saved_model
    from tensorflow.python.saved_model import save_options
    from tensorflow.python.saved_model.load import load
    from tensorflow.python.saved_model.loader_impl import parse_saved_model_with_debug_info

    version = "saved_model_v2"
    try:
        saved_model.save(model, saved_model_dir, options=save_options.SaveOptions(save_debug_info=True))
    except:
        return "trackable_object"
    saved_model_proto, _ = parse_saved_model_with_debug_info(saved_model_dir)
    saved_model_version = saved_model_proto.saved_model_schema_version
    if saved_model_version == 0:
        return "saved_model_v1"
    if saved_model_version not in [1, 2]:
        raise ValueError("SavedModel file format({0}) is not supported".format(saved_model_version))
    return version


def _get_graph_from_saved_model_v1(model):
    """The version 1 function that get graph from saved_model.

    Args:
        model (string or tf.keras.Model): model path or tf.keras.Model object.

    Returns:
        graph_def (tf.compat.v1.Session): tf.compat.v1.Session object.
        inputs (list of string): validated input names.
        outputs (list of string): validated output names.
    """
    from tensorflow.lite.python.convert_saved_model import get_inputs_outputs, get_meta_graph_def, get_signature_def
    from tensorflow.python.client import session
    from tensorflow.python.framework import ops
    from tensorflow.python.saved_model import constants, signature_constants, tag_constants

    saved_model_tags = set([tag_constants.SERVING])
    signature_key = signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY

    meta_graph = get_meta_graph_def(model, saved_model_tags)
    signature_def = get_signature_def(meta_graph, signature_key)
    inputs, outputs = get_inputs_outputs(signature_def)  # pragma: no cover
    # Check SavedModel for assets directory.
    collection_def = meta_graph.collection_def
    if constants.ASSETS_KEY in collection_def:  # pragma: no cover
        raise ValueError("SavedModels with assets/ directory are not supported.")

    from tensorflow.compat.v1 import graph_util as tf_graph_util
    from tensorflow.python.saved_model import loader

    graph = ops.Graph()
    import tensorflow as tf

    with session.Session(graph=graph) as sess:
        loader.load(sess, meta_graph.meta_info_def.tags, model)
        sess.run(tf.compat.v1.global_variables_initializer())
        sess.run(tf.compat.v1.tables_initializer())
        output_nodes = list(set([output.split(":")[0] for output in outputs]))
        node_ops = [node.op for node in graph.as_graph_def().node]
        if "MakeIterator" in node_ops:  # pragma: no cover
            output_nodes.append("MakeIterator")
        table_ops = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.TABLE_INITIALIZERS)
        # For table initialization
        for table_op in table_ops:  # pragma: no cover
            output_nodes.append(table_op.name)
        if len(table_ops) > 0:  # pragma: no cover
            output_nodes.append("init_all_tables")
        graph_def = tf_graph_util.convert_variables_to_constants(sess, graph.as_graph_def(), output_nodes)
    return graph_def, inputs, outputs


def try_loading_keras(model, input_tensor_names, output_tensor_names):  # pragma: no cover
    """Try different ways of loading keras models.

    Args:
        model (string or tf.keras.Model): model path or tf.keras.Model object.
        input_tensor_names (list of string): input tensor names of the model.
        output_tensor_names (list of string): output tensor names of the model.

    Returns:
        graph_def (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_names (list of string): validated input names.
        output_names (list of string): validated output names.
    """
    temp_dir = tempfile.mkdtemp()
    if not isinstance(model, tf.keras.Model):
        model = tf.keras.models.load_model(model)
    keras_format = _check_keras_format(model, temp_dir)

    if keras_format == "saved_model_v2":
        try:
            graph_def, input_names, output_names = _get_graph_from_saved_model_v2(
                temp_dir, input_tensor_names, output_tensor_names
            )
            if "_FusedBatchNormEx" in [node.op for node in graph_def.node]:
                keras_format = "trackable_object"
        except:
            keras_format = "trackable_object"

    if keras_format == "trackable_object":
        try:
            graph_def, input_names, output_names = _get_graph_from_original_keras_v2(model)
        except:
            keras_format = "saved_model_v1"

    if keras_format == "saved_model_v1":  # pragma: no cover
        try:
            tf.keras.backend.set_learning_phase(0)
            graph_def, input_names, output_names = _get_graph_from_saved_model_v1(model)
        except:
            raise ValueError("Not supported keras model type...")

    shutil.rmtree(temp_dir, True)
    return graph_def, input_names, output_names


def keras_session(model, input_tensor_names, output_tensor_names, **kwargs):
    """Build session with keras model.

    Args:
        model (string or tf.keras.Model): model path or tf.keras.Model object.
        input_tensor_names (list of string): input_tensor_names of model.
        output_tensor_names (list of string): output_tensor_names of model.

    Returns:
        sess (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_tensor_names (list of string): validated input_tensor_names.
        output_tensor_names (list of string): validated output_tensor_names.
    """
    if tf.version.VERSION > "2.1.0":
        try:
            graph_def, input_names, output_names = _get_graph_from_saved_model_v3(
                model, input_tensor_names, output_tensor_names
            )
        except:
            graph_def, input_names, output_names = try_loading_keras(model, input_tensor_names, output_tensor_names)
    # tensorflow 1.x use v1 convert method
    else:
        tf.keras.backend.set_learning_phase(0)
        graph_def, input_names, output_names = _get_graph_from_saved_model_v1(model)

    return graph_def_session(graph_def, input_names, output_names, **kwargs)


def slim_session(model, input_tensor_names, output_tensor_names, **kwargs):  # pragma: no cover
    """Build session with slim model.

    Args:
        model (string): model path.
        input_tensor_names (list of string): input_tensor_names of model.
        output_tensor_names (list of string): output_tensor_names of model.

    Returns:
        sess (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_tensor_names (list of string): validated input_tensor_names.
        output_tensor_names (list of string): validated output_tensor_names.
    """
    assert version1_lt_version2(tf.version.VERSION, "2.0.0"), "slim model only used in tensorflow 1.x"
    from neural_compressor.tensorflow.utils.utility import TFSlimNetsFactory

    factory = TFSlimNetsFactory()
    assert "name" in kwargs, "model name should be set in slim checkpoint...."
    assert kwargs["name"] in factory.default_slim_models, "only support topology {}".format(factory.default_slim_models)
    net = copy.deepcopy(factory.networks_map[kwargs["name"]])
    model_func = net.pop("model")
    arg_scope = net.pop("arg_scope")()
    inputs_shape = net.pop("input_shape")
    kwargs = net
    import tf_slim as slim

    with tf.Graph().as_default():
        images = tf.compat.v1.placeholder(name="input", dtype=tf.float32, shape=inputs_shape)
        with tf.compat.v1.Session() as sess:
            with slim.arg_scope(arg_scope) as scope:  # pylint: disable=not-context-manager
                model_func(images, is_training=False, **kwargs)
            graph_def = sess.graph.as_graph_def()
            output_tensor_names = output_tensor_names if len(output_tensor_names) > 0 else [graph_def.node[-1].name]

            from tensorflow.python.tools.freeze_graph import freeze_graph_with_def_protos

            graph_def = freeze_graph_with_def_protos(
                input_graph_def=graph_def,
                input_saver_def=None,
                input_checkpoint=model,
                output_node_names=",".join(output_tensor_names),
                restore_op_name="save/restore_all",
                filename_tensor_name="save/Const:0",
                output_graph="",
                clear_devices=True,
                initializer_nodes="",
            )

    return graph_def_session(graph_def, ["input"], output_tensor_names, **kwargs)


def checkpoint_session(model, input_tensor_names, output_tensor_names, **kwargs):
    """Build session with ckpt model.

    Args:
        model (string): model path.
        input_tensor_names (list of string): input_tensor_names of model.
        output_tensor_names (list of string): validated output_tensor_names of model.

    Returns:
        sess (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_tensor_names (list of string): validated input_tensor_names.
        output_tensor_names (list of string): validated output_tensor_names.
    """
    assert (
        output_tensor_names is not None and len(output_tensor_names) > 0
    ), "outputs should not be None of checkpoint...."

    ckpt_prefix = [os.path.splitext(i)[0] for i in os.listdir(model) if i.endswith(".meta")][0]

    config = tf.compat.v1.ConfigProto()
    config.use_per_session_threads = 1
    config.inter_op_parallelism_threads = 1
    graph = tf.Graph()
    sess = tf.compat.v1.Session(graph=graph, config=config)
    if version1_lt_version2(tf.version.VERSION, "2.0.0"):  # pragma: no cover
        from tensorflow._api.v1.config import experimental

        list_physical_devices = experimental.list_physical_devices
    else:
        list_physical_devices = tf.config.list_physical_devices

    with graph.as_default():
        device = kwargs.get("device")
        if device == "cpu":
            cpus = list_physical_devices("CPU")
            node_device = cpus[0].name.replace("physical_device:", "")
            with graph.device(node_device):
                saver = tf.compat.v1.train.import_meta_graph(
                    os.path.join(model, ckpt_prefix + ".meta"), clear_devices=True
                )
        else:  # pragma: no cover
            saver = tf.compat.v1.train.import_meta_graph(os.path.join(model, ckpt_prefix + ".meta"), clear_devices=True)

        sess.run(tf.compat.v1.global_variables_initializer())
        saver.restore(sess, os.path.join(model, ckpt_prefix))

    from neural_compressor.tensorflow.quantization.utils.utility import get_input_output_node_names

    if validate_graph_node(sess.graph.as_graph_def(), tensor_to_node(input_tensor_names)):
        input_tensor_names = input_tensor_names
    else:
        input_tensor_names, _ = get_input_output_node_names(sess.graph.as_graph_def())
    return sess, input_tensor_names, output_tensor_names


def estimator_session(model, input_tensor_names, output_tensor_names, **kwargs):  # pragma: no cover
    """Build session with estimator model.

    Args:
        model (tf.estimator.Estimator): tf.estimator.Estimator object.
        input_tensor_names (list of string): input_tensor_names of model.
        output_tensor_names (list of string): output_tensor_names of model.
        kwargs (dict): other required parameters, like input_fn.

    Returns:
        sess (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_tensor_names (list of string): validated input_tensor_names.
        output_tensor_names (list of string): validated output_tensor_names.
    """
    assert "input_fn" in kwargs, "input func should be supplied for estimator session...."
    with tf.Graph().as_default() as g:
        features, input_hooks = model._get_features_from_input_fn(kwargs["input_fn"], tf.estimator.ModeKeys.PREDICT)
        estimator_spec = model._call_model_fn(features, None, tf.estimator.ModeKeys.PREDICT, model.config)

        if len(output_tensor_names) == 0:
            outputs = (
                [tensor.name for tensor in estimator_spec.predictions.values()]
                if isinstance(estimator_spec.predictions, dict)
                else [estimator_spec.predictions.name]
            )
        else:
            outputs = output_tensor_names

        logger.info("Estimator output tensor names are {}.".format(outputs))
        with tf.compat.v1.Session(graph=g) as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            # Freezing a graph requires output_node_names, which can be found in
            # estimator_spec.predictions that contains prediction tensors as a
            # dictionary
            # When a model uses Iterator, we need to have 'MakeIterator' (default
            # name used by TF) in the output_node_names as well.
            output_nodes = list(set([output.split(":")[0] for output in outputs]))
            if "MakeIterator" in [node.op for node in g.as_graph_def().node]:
                output_nodes.append("MakeIterator")

            graph_def = tf.compat.v1.graph_util.convert_variables_to_constants(sess, g.as_graph_def(), output_nodes)

        return graph_def_session(graph_def, input_tensor_names, outputs, **kwargs)


def saved_model_session(model, input_tensor_names, output_tensor_names, **kwargs):
    """Build session with saved model.

    Args:
        model (string): model path.
        input_tensor_names (list of string): input_tensor_names of model.
        output_tensor_names (list of string): output_tensor_names of model.

    Returns:
        sess (tf.compat.v1.Session): tf.compat.v1.Session object.
        input_tensor_names (list of string): validated input_tensor_names.
        output_tensor_names (list of string): validated output_tensor_names.
    """
    try:
        graph_def, input_names, output_names = _get_graph_from_saved_model_v3(
            model, input_tensor_names, output_tensor_names
        )
    except:
        try:
            graph_def, input_names, output_names = _get_graph_from_saved_model_v2(
                model, input_tensor_names, output_tensor_names
            )
        except:
            graph_def, input_names, output_names = _get_graph_from_saved_model_v1(model)

    assert graph_def is not None, "Can not parse the saved model..."

    return graph_def_session(graph_def, input_names, output_names, **kwargs)


# it's necessary that a session with input output tensors to run the model
SESSIONS = {
    "frozen_pb": frozen_pb_session,
    "graph_def": graph_def_session,
    "graph": graph_session,
    "saved_model": saved_model_session,
    "llm_saved_model": saved_model_session,
    "keras": keras_session,
    "checkpoint": checkpoint_session,
    "estimator": estimator_session,
    "slim": slim_session,
}


class BaseModel:
    """Base class of all neural_compressor.model, will play graph role."""

    def __init__(self, model, **kwargs):
        """Initialize a BaseModel.

        Args:
            model (object): raw model format. For Tensorflow model, could be path to frozen pb file,
                path to ckpt or savedmodel folder, loaded estimator/graph_def/graph/keras model object.
        """
        self.component = None

    @property
    def model(self):
        """Return model itself."""
        raise NotImplementedError

    @property
    def graph_info(self):
        """Return a dict with content 'Node: Node_type'."""
        raise NotImplementedError

    @abstractmethod
    def save(self, root, *args, **kwargs):
        """Abstract method of model saving."""
        raise NotImplementedError

    @abstractmethod
    def export(
        self,
        save_path: str,
        conf,
    ):
        """Abstract method of model conversion to ONNX."""
        raise NotImplementedError

    @abstractmethod
    def framework(self):
        """Abstract method of model framework."""
        raise NotImplementedError


class TensorflowBaseModel(BaseModel):
    """Build Tensorflow Base Model."""

    def __init__(self, model, **kwargs):
        """Initialize a Tensorflow model.

        Args:
            model (string or tensorflow model object): model path or model object.
        """
        self._model = model
        self._name = ""
        self._weights = None
        self.kwargs = kwargs
        self._graph_info = {}
        self._input_tensor_names = []
        self._output_tensor_names = []
        self._model_type = ""
        self._sess = None
        self._iter_op = None
        self._workspace_path = ""
        self._q_config = None
        self._model_path = None if not isinstance(model, str) else model

    @property
    def model_path(self):
        """Return model path."""
        return self._model_path

    @model_path.setter
    def model_path(self, path):
        """Set model path."""
        self._model_path = path

    def framework(self):
        """Return framework."""
        return "tensorflow"

    @property
    def name(self):
        """Renturn name."""
        return self._name

    @name.setter
    def name(self, name):
        """Set name."""
        self.kwargs.update({"name": name})
        self._name = name

    @property
    def weights(self):
        """Return weights."""
        return self._weights

    @weights.setter
    def weights(self, new_weights):
        """Set weights."""
        self._weights = new_weights

    @property
    def q_config(self):
        """Return q_config."""
        return self._q_config

    @q_config.setter
    def q_config(self, q_config):
        """Set q_config."""
        self._q_config = q_config

    @property
    def workspace_path(self):
        """Return workspace path."""
        return self._workspace_path

    @workspace_path.setter
    def workspace_path(self, path):
        """Set workspace path."""
        self._workspace_path = path

    @property
    def model_type(self):
        """Return model type."""
        return self._model_type

    @model_type.setter
    def model_type(self, model_type):
        """Set model type."""
        assert model_type in SESSIONS, "model type not supported...."
        self._model_type = model_type

    @property
    def model(self):
        """Return model itself."""
        return self.graph

    @property
    def graph_def(self):
        """Return graph definition."""
        return self.graph.as_graph_def()

    @property
    def graph_info(self):
        """Return graph info."""
        self._graph_info = {}
        for node in self.graph_def.node:
            self._graph_info[node.name] = node.op
        return self._graph_info

    @property
    def sess(self):
        """Return Session object."""
        if self._sess is None:
            self._load_sess(self._model, **self.kwargs)
        return self._sess

    @property
    def graph(self):
        """Return model graph."""
        return self.sess.graph

    @graph_def.setter
    def graph_def(self, graph_def):
        """Set graph definition."""
        if self._sess is not None:
            self._sess.close()
        output_sess = SESSIONS["graph_def"](
            graph_def, self._input_tensor_names, self._output_tensor_names, **self.kwargs
        )

        self._sess = output_sess[0]
        self._input_tensor_names = output_sess[1]
        self._output_tensor_names = output_sess[2]
        self.model_type = "graph_def"

    def _load_sess(self, model, **kwargs):
        if self.name:
            kwargs.update({"name": self.name})
        # assert self.model_type, 'model type not set....'
        output_sess = SESSIONS[self.model_type](model, self._input_tensor_names, self._output_tensor_names, **kwargs)
        self._sess = output_sess[0]
        self._input_tensor_names = output_sess[1]
        self._output_tensor_names = output_sess[2]

        tf.compat.v1.get_variable_scope().reuse_variables()
        return self._sess

    @property
    def iter_op(self):
        """Return model iter op list."""
        self._iter_op = []
        if self._sess is None:  # pragma: no cover
            self._load_sess(self._model, **self.kwargs)
        op_list = [node.op for node in self._sess.graph.as_graph_def().node]
        if "MakeIterator" in op_list:  # pragma: no cover
            self._iter_op.append(self._sess.graph.get_operation_by_name("MakeIterator"))
        return self._iter_op

    @property
    def input_tensor_names(self):
        """Return input tensor names."""
        if self._sess is None:
            self._load_sess(self._model, **self.kwargs)
        return copy.deepcopy(self._input_tensor_names)

    @input_tensor_names.setter
    def input_tensor_names(self, tensor_names):
        """Set input tensor names."""
        if len(tensor_names) == 0:  # pragma: no cover
            logger.warning("Input tensor names is empty.")
            return
        if self._sess is not None:
            assert validate_graph_node(
                self.graph_def, tensor_to_node(tensor_names)
            ), "tensor names {} not in graph".format(tensor_names)
        self._input_tensor_names = tensor_names

    @property
    def output_tensor_names(self):
        """Return output tensor names."""
        if len(self._output_tensor_names) == 0:
            self._load_sess(self._model, **self.kwargs)
        return copy.deepcopy(self._output_tensor_names)

    @output_tensor_names.setter
    def output_tensor_names(self, tensor_names):
        """Set output tensor names."""
        if len(tensor_names) == 0:  # pragma: no cover
            logger.warning("Output tensor names should not be empty.")
            return
        if self._sess is not None:
            assert validate_graph_node(
                self.graph_def, tensor_to_node(tensor_names)
            ), "tensor names {} not in graph".format(tensor_names)
        self._output_tensor_names = tensor_names

    # input/output node names and input/output tensor
    # come from input/output tensor names, so do not support assign these values
    @property
    def input_node_names(self):
        """Return input node names."""
        return copy.deepcopy(tensor_to_node(self.input_tensor_names))

    @property
    def output_node_names(self):
        """Return output node names."""
        output_node_names = tensor_to_node(self.output_tensor_names)
        iter_op_list = self.iter_op
        if iter_op_list != []:  # pragma: no cover
            output_node_names += [iter_op.name for iter_op in iter_op_list]
        return copy.deepcopy(output_node_names)

    @property
    def input_tensor(self):
        """Return input tensor."""
        from neural_compressor.tensorflow.quantization.utils.utility import get_tensor_by_name

        return [get_tensor_by_name(self.graph, x) for x in self.input_tensor_names]

    @property
    def output_tensor(self):
        """Return output tensor."""
        from neural_compressor.tensorflow.quantization.utils.utility import get_tensor_by_name

        return [get_tensor_by_name(self.graph, x) for x in self.output_tensor_names]

    def save(self, root=None):
        """Save Tensorflow model."""
        if not root:  # pragma: no cover
            root = DEFAULT_WORKSPACE + "/save.pb"
        root = os.path.abspath(os.path.expanduser(root))
        # if not have suffix, default append .pb
        os.makedirs(os.path.dirname(root), exist_ok=True)
        pb_file = root if os.path.split(root)[-1].endswith(".pb") else root + ".pb"
        f = tf.io.gfile.GFile(pb_file, "wb")
        f.write(self.graph_def.SerializeToString())
        logger.info("Save quantized model to {}.".format(pb_file))


class TensorflowSavedModelModel(TensorflowBaseModel):
    """Build Tensorflow saved model."""

    def __init__(self, model, **kwargs):
        """Initialize a Tensorflow model.

        Args:
            model (string or tensorflow model object): model path or model object.
        """
        super(TensorflowSavedModelModel, self).__init__(model, **kwargs)
        self._auto_trackable = None

    def get_all_weight_names(self):
        """Get weight names of model.

        Returns:
            list: weight names list.
        """
        import tensorflow as tf

        names = []
        for index, layer in enumerate(tf.keras.models.load_model(self._model).layers):
            if len(layer.weights):
                names.append(index)
        return names

    def update_weights(self, tensor_name, new_tensor):
        """Update model weights."""
        pass

    def get_weight(self, tensor_name):
        """Return model weight with a given tensor name.

        Args:
            tensor_name (str): name of a tensor.
        """
        return self.weights[tensor_name]

    @property
    def model(self):
        """Return model in AutoTrackable object."""
        if self._auto_trackable:  # pragma: no cover
            return self._auto_trackable

        root = os.path.abspath(os.path.expanduser(DEFAULT_WORKSPACE))
        root += str(time.time())
        if os.path.exists(root):  # pragma: no cover
            shutil.rmtree(root)
        os.makedirs(root, exist_ok=True)
        if not self._sess:
            self._load_sess(self._model, **self.kwargs)
        _, builder = self.build_saved_model(root)
        builder.save()
        model = tf.saved_model.load(root)
        shutil.rmtree(root)
        self._auto_trackable = model
        return model

    @model.setter
    def model(self, input_model):
        """Set model in AutoTrackable object."""
        self._auto_trackable = input_model

    def build_saved_model(self, root=None):
        """Build Tensorflow saved model.

        Args:
            root (str, optional): path to saved model. Defaults to None.

        Returns:
            root (str): path to saved model.
            builder (tf.compat.v1.saved_model.builder.SavedModelBuilder): builds
                the SavedModel protocol buffer and saves variables and assets.
        """
        if not root:  # pragma: no cover
            root = DEFAULT_WORKSPACE
        root = os.path.abspath(os.path.expanduser(root))
        if os.path.exists(root):
            import shutil

            shutil.rmtree(root)

        os.makedirs(root, exist_ok=True)

        from tensorflow.python.saved_model import signature_constants, tag_constants

        from neural_compressor.tensorflow.quantization.utils.utility import get_tensor_by_name

        builder = tf.compat.v1.saved_model.builder.SavedModelBuilder(root)
        sigs = {}
        with tf.compat.v1.Session(graph=tf.Graph()) as sess:
            # (TODO) not directly use self._sess.graph, use self.graph
            tf.import_graph_def(self.graph.as_graph_def(), name="")
            g = tf.compat.v1.get_default_graph()
            inp = [get_tensor_by_name(g, x) for x in self._input_tensor_names]
            out = [get_tensor_by_name(g, x) for x in self._output_tensor_names]
            sigs[signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY] = (
                tf.compat.v1.saved_model.signature_def_utils.predict_signature_def(
                    {k: v for k, v in zip(self._input_tensor_names, inp)},
                    {k: v for k, v in zip(self._output_tensor_names, out)},
                )
            )
            builder.add_meta_graph_and_variables(sess, [tag_constants.SERVING], signature_def_map=sigs)
        return root, builder

    def save(self, root=None):
        """Save Tensorflow model."""
        root, builder = self.build_saved_model(root)
        builder.save()
        logger.info("Save quantized model to {}.".format(root))


class TensorflowLLMModel(TensorflowSavedModelModel):
    """The class Tensorflow saved model whose GraphDef exceeding maximum protobuf size of 2GB."""

    def __init__(self, model, **kwargs):
        """Initialize a Tensorflow model.

        Args:
            model (string or tensorflow model object): model path or model object.
        """
        super(TensorflowLLMModel, self).__init__(model, **kwargs)

        self._model_path = self.kwargs.get("model_path", None)
        self._weight_name_mapping = self.kwargs.get("weight_name_mapping", None)
        self._sq_weight_scale_dict = self.kwargs.get("sq_weight_scale_dict", None)
        self._weight_tensor_minmax_dict = {}
        self._model_type = "llm_saved_model"

        from neural_compressor.tensorflow.quantization.utils.utility import parse_saved_model

        (
            self._graph_def,
            self._saved_model,
            self.func,
            self.frozen_func,
            self._input_tensor_names,
            self._output_tensor_names,
        ) = parse_saved_model(model)

    @property
    def model_path(self):
        """Return model path.

        The model path in this class is used as a temp path for intermediate model
        """
        return self._model_path

    @model_path.setter
    def model_path(self, path):
        """Set model path.

        The model path in this class is used as a temp path for intermediate model
        """
        self.kwargs.update({"model_path": path})
        self._model_path = path

    @property
    def graph_def(self):
        """Return graph_def."""
        return self._graph_def

    @graph_def.setter
    def graph_def(self, graph_def):
        """Set graph definition."""
        self._graph_def = graph_def
        # the attributes of some nodes can't be correctly read if don't import the graph_def
        tf.import_graph_def(self._graph_def, name="")

    @property
    def model(self):
        """Return model in AutoTrackable Format."""
        if self._sq_weight_scale_dict:
            self.adjust_weight(self.graph_def)
        if not self._auto_trackable:
            self._auto_trackable = tf.saved_model.load(self._model)
        return self._auto_trackable

    @property
    def weight_name_mapping(self):
        """Return weight_name_mapping function."""
        if not self._weight_name_mapping:  # pragma: no cover
            self._weight_name_mapping = self.kwargs.get("weight_name_mapping", None)
        assert self._weight_name_mapping is not None, "weight_name_mapping should not be None!"
        return self._weight_name_mapping

    @weight_name_mapping.setter
    def weight_name_mapping(self, weight_name_mapping):
        """Set weight_name_mapping function."""
        self.kwargs.update({"weight_name_mapping": weight_name_mapping})
        self._weight_name_mapping = weight_name_mapping

    @property
    def sq_weight_scale_dict(self):
        """Return dict of weight scaler for smooth quantization."""
        if not self._sq_weight_scale_dict:  # pragma: no cover
            self._sq_weight_scale_dict = self.kwargs.get("sq_weight_scale_dict", None)
        assert self._weight_name_mapping is not None, "sq_weight_scale_dict should not be None!"
        return self._sq_weight_scale_dict

    @sq_weight_scale_dict.setter
    def sq_weight_scale_dict(self, sq_weight_scale_dict):
        """Set dict of weight scaler for smooth quantization."""
        self.kwargs.update({"sq_weight_scale_dict": sq_weight_scale_dict})
        self._sq_weight_scale_dict = sq_weight_scale_dict

    @property
    def weight_tensor_minmax_dict(self):
        """Return dict of weight scaler for smooth quantization."""
        return self._weight_tensor_minmax_dict

    @property
    def input_tensor_names(self):
        """Return input tensor names."""
        return copy.deepcopy(self._input_tensor_names)

    @input_tensor_names.setter
    def input_tensor_names(self, tensor_names):
        """Set input tensor names."""
        if len(tensor_names) == 0:  # pragma: no cover
            logger.warning("Input tensor names is empty.")
            return

        assert validate_graph_node(
            self._graph_def, tensor_to_node(tensor_names)
        ), "tensor names {} not in graph".format(tensor_names)
        self._input_tensor_names = tensor_names

    @property
    def output_tensor_names(self):
        """Return output tensor names."""
        return copy.deepcopy(self._output_tensor_names)

    @output_tensor_names.setter
    def output_tensor_names(self, tensor_names):
        """Set output tensor names."""
        if len(tensor_names) == 0:  # pragma: no cover
            logger.warning("Output tensor names is empty.")
            return
        if self._graph_def is not None:
            assert validate_graph_node(
                self.graph_def, tensor_to_node(tensor_names)
            ), "tensor names {} not in graph".format(tensor_names)
        self._output_tensor_names = tensor_names

    @property
    def output_node_names(self):
        """Return output node names."""
        output_node_names = tensor_to_node(self.output_tensor_names)
        return copy.deepcopy(output_node_names)

    def adjust_weight(self, graph_def):
        """Adjust weight of LLM saved_model by scale."""
        from tensorflow.python.saved_model import load, tag_constants

        from neural_compressor.tensorflow.quantization.utils.utility import reconstruct_saved_model

        reconstruct_saved_model(graph_def, self.func, self.frozen_func, self._saved_model, self.model_path)
        model = load.load(self.model_path, [tag_constants.SERVING])

        for idx, weight_tensor in enumerate(model.variables):
            parsed_weight_name = self.weight_name_mapping(weight_tensor.name)
            if parsed_weight_name in self.sq_weight_scale_dict:
                if len(weight_tensor.shape) == 4:  # pragma: no cover
                    shape_parm = [0, 1, 3, 2]
                elif len(weight_tensor.shape) == 2:
                    shape_parm = [1, 0]
                weight_array = np.transpose(weight_tensor, shape_parm)
                weight_array *= self.sq_weight_scale_dict[parsed_weight_name]
                weight_array = np.transpose(weight_array, shape_parm)
                tf.compat.v1.assign(model.variables[idx], weight_array)
            else:
                weight_array = weight_tensor

            if parsed_weight_name not in self._weight_tensor_minmax_dict:
                self._weight_tensor_minmax_dict[parsed_weight_name] = [np.min(weight_array), np.max(weight_array)]
        self._auto_trackable = model

    def save(self, root=None):
        """Save the model to the root path."""
        import shutil

        from neural_compressor.tensorflow.quantization.utils.utility import parse_saved_model, reconstruct_saved_model

        if not root:  # pragma: no cover
            root = DEFAULT_WORKSPACE
        root = os.path.abspath(os.path.expanduser(root))
        if os.path.exists(root):  # pragma: no cover
            shutil.rmtree(root)
        os.makedirs(root, exist_ok=True)

        self.adjust_weight(self._graph_def)
        graph_def, _saved_model, func, frozen_func, _, _ = parse_saved_model(self._auto_trackable)
        reconstruct_saved_model(graph_def, func, frozen_func, _saved_model, root)
        logger.info("Save quantized model to {}.".format(root))
        # delete the LLM file saved in this temporary path
        shutil.rmtree(self.model_path, ignore_errors=True)


class TensorflowCheckpointModel(TensorflowBaseModel):
    """Build Tensorflow checkpoint model."""

    @property
    def graph_def(self):
        """Return graph definition."""
        if self.model_type == "graph_def":  # pragma: no cover
            return self.sess.graph.as_graph_def()
        from tensorflow.compat.v1 import graph_util

        from neural_compressor.tensorflow.quantization.utils.utility import _parse_ckpt_bn_input

        graph_def = self.sess.graph.as_graph_def()
        graph_def = _parse_ckpt_bn_input(graph_def)
        return graph_util.convert_variables_to_constants(
            sess=self._sess, input_graph_def=graph_def, output_node_names=self.output_node_names
        )

    @graph_def.setter
    def graph_def(self, graph_def):
        """Set graph definition."""
        if self._sess is not None:
            self._sess.close()
        output_sess = SESSIONS["graph_def"](
            graph_def, self._input_tensor_names, self._output_tensor_names, **self.kwargs
        )
        self._sess = output_sess[0]
        self._input_tensor_names = output_sess[1]
        self._output_tensor_names = output_sess[2]
        self.model_type = "graph_def"

    @property
    def model(self):
        """Return the model itself to avoid the initialization issue."""
        return self


class KerasModel(BaseModel):
    """Build Keras model."""

    def __init__(self, model, **kwargs):
        """Initialize a Keras model.

        Args:
            model (string or keras model object): model path or model object.
        """
        self.component = None
        self._model = model
        if not isinstance(model, tf.keras.Model):
            self._model_object = tf.keras.models.load_model(self._model)
        else:
            self._model_object = self._model
        self._q_config = None

    @property
    def q_config(self):
        """Return q_config."""
        return self._q_config

    @q_config.setter
    def q_config(self, q_config):
        """Set q_config."""
        self._q_config = q_config

    @property
    def model(self):
        """Return model itself."""
        return self._model_object

    @property
    def graph_info(self):
        """Return graph info."""
        # (TODO) get the graph info
        return None

    @abstractmethod
    def save(self, root, *args, **kwargs):
        """Save Keras model."""
        self._model_object.save(root)
        logger.info("Save quantized model to {}.".format(root))

    @property
    def input_node_names(self):
        """Return input node names."""
        names = (
            self.model.input_names
            if version1_lt_version2(tf.version.VERSION, "2.16.1")
            else [tensor.name for tensor in self.model.inputs]
        )

        return names

    @property
    def output_node_names(self):
        """Return output node names."""
        names = (
            self.model.output_names
            if version1_lt_version2(tf.version.VERSION, "2.16.1")
            else [tensor.name for tensor in self.model.outputs]
        )

        return names


TENSORFLOW_MODELS = {
    "frozen_pb": TensorflowBaseModel,
    "graph_def": TensorflowBaseModel,
    "graph": TensorflowBaseModel,
    "checkpoint": TensorflowCheckpointModel,
    "estimator": TensorflowBaseModel,
    "slim": TensorflowBaseModel,
    "saved_model": TensorflowSavedModelModel,
    "AutoTrackable": TensorflowSavedModelModel,
    "llm_saved_model": TensorflowLLMModel,
    "keras": KerasModel,
}


class TensorflowModel(object):
    """A wrapper to construct a Tensorflow Model."""

    def __new__(cls, model_type, root, **kwargs):
        """Create a new instance object of TensorflowModel.

        Args:
            model_type (str): model type.
            root (str): model path.

        Returns:
            tensorflow model object: tensorflow model.
        """
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        model = TENSORFLOW_MODELS[model_type](root, **kwargs)
        model.model_type = model_type
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Merge duplicated QDQ patterns Graph Rewriter."""

from tensorflow.core.framework import node_def_pb2

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class MergeDuplicatedQDQOptimizer(GraphRewriterBase):
    """Merge duplicated QDQ patterns."""

    @dump_elapsed_time("Pass MergeDuplicatedQDQOptimizer")
    def do_transformation(self):
        """Execute the merge."""
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model
        graph_info = cur_graph.parse_graph()

        patterns = [["QuantizeV2"], ["Dequantize"]]
        matched_nodes = cur_graph.query_fusion_pattern_nodes(patterns)

        quantizev2_input_map = {}
        dequantize_map = {}
        for i in matched_nodes:
            quantizev2_input_node_name = graph_info[i[0]].node.input[0]
            if quantizev2_input_node_name in quantizev2_input_map:
                quantizev2_input_map[quantizev2_input_node_name].append(graph_info[i[0]].node)
                dequantize_map[quantizev2_input_node_name].append(graph_info[i[1]].node)
            else:
                quantizev2_input_map[quantizev2_input_node_name] = [graph_info[i[0]].node]
                dequantize_map[quantizev2_input_node_name] = [graph_info[i[1]].node]

        for input_map_node_name, quantizev2_nodes in quantizev2_input_map.items():
            if input_map_node_name not in cur_graph.node_name_details:
                continue

            # the nodes share the the same input and have the same QuantizeV2 op type
            if len(quantizev2_nodes) > 1:
                # merge the duplicated QuantizeV2 op.
                new_quantize_node = quantizev2_nodes[0]
                new_dequantize_node = dequantize_map[input_map_node_name][0]

                do_merge = True
                for i in quantizev2_nodes:
                    if i.name != new_quantize_node.name and i.attr["T"].type != new_quantize_node.attr["T"].type:
                        do_merge = False
                        break

                if not do_merge:
                    continue

                # set the new QuantizeV2 node as the only output of the parent node
                for i in quantizev2_nodes:
                    if i.name != new_quantize_node.name:
                        cur_graph.node_name_details[input_map_node_name].outputs.remove(i.name)

                # set the new QuantizeV2 node as all the other input of the Dequantize nodes
                for i in dequantize_map[input_map_node_name]:
                    if i.name != new_dequantize_node.name:
                        cur_graph.node_name_details[i.name].node.ClearField("input")
                        cur_graph.node_name_details[i.name].node.input.extend(
                            [new_quantize_node.name, new_quantize_node.name + ":1", new_quantize_node.name + ":2"]
                        )

                # remove the duplicated quantized nodes
                for i in quantizev2_nodes:
                    if i.name != new_quantize_node.name:
                        # remove quantize min node
                        cur_graph.remove_node(cur_graph.node_name_details[i.name].node.input[1])
                        # remove quantize max node
                        cur_graph.remove_node(cur_graph.node_name_details[i.name].node.input[2])
                        # remove quantize node
                        cur_graph.remove_node(i.name)

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Rename FusedBatchNorm op to FusedBatchNormV2 Graph Rewriter."""

import math

import numpy as np
from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class RenameBatchNormOptimizer(GraphRewriterBase):
    """Rename FusedBatchNorm op to FusedBatchNormV2."""

    @dump_elapsed_time("Pass RenameBatchNormOptimizer")
    def do_transformation(self):
        """Rename FusedBatchNorm op to FusedBatchNormV2.

        This pass is needed for bf16 conversion. Due to TensorFlow historical reason,
        FusedBatchNorm is not a bf16 op but FusedBatchNormV2 is. As the latter is compatible
        with the former, changing FusedBatchNorm op to FusedBatchNormV2 op will be able to
        convert to bf16 op on the platforms supporting VNNI_BF16 and AMX instructions.

        Returns:
          Modified graph with BN ops renamed.

        Raises:
          ValueError: If the graph is badly formed.
        """
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model
        graph_details = cur_graph.parse_graph()

        for _, v in graph_details.items():
            # for node in cur_graph.graph.node:
            if v.node.op == "FusedBatchNorm" or v.node.op == "FusedBatchNormV2":
                v.node.op = "FusedBatchNormV3"
                v.node.attr["U"].CopyFrom(v.node.attr["T"])

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The quantization classes for Keras."""

import copy
import json
import os
from collections import OrderedDict, UserDict
from typing import Callable, Dict

import numpy as np
import tensorflow as tf
import yaml

from neural_compressor.common import logger
from neural_compressor.common.utils import DEFAULT_WORKSPACE
from neural_compressor.tensorflow.keras.layers import (
    QAvgPool2D,
    QConv2D,
    QDense,
    QDepthwiseConv2D,
    QMaxPool2D,
    QSeparableConv2D,
)
from neural_compressor.tensorflow.quantization.config import StaticQuantConfig
from neural_compressor.tensorflow.utils import deep_get, dump_elapsed_time, version1_gte_version2


class KerasAdaptor:
    """The keras class of framework adaptor layer."""

    supported_op = [
        "Conv2D",
        "Dense",
        "SeparableConv2D",
        "DepthwiseConv2D",
        "AveragePooling2D",
        "MaxPooling2D",
        "AvgPool2D",
        "MaxPool2D",
    ]

    custom_layers = {
        "QConv2D": QConv2D,
        "QDepthwiseConv2D": QDepthwiseConv2D,
        "QSeparableConv2D": QSeparableConv2D,
        "QDense": QDense,
        "QMaxPool2D": QMaxPool2D,
        "QAvgPool2D": QAvgPool2D,
        "QMaxPooling2D": QMaxPool2D,
        "QAveragePooling2D": QAvgPool2D,
    }

    def __init__(self, framework_specific_info):
        """Initialize the KerasAdaptor class with framework specific information."""
        self.framework_specific_info = framework_specific_info
        self.approach = deep_get(self.framework_specific_info, "approach", False)
        self.quantize_config = {"op_wise_config": {}}
        self.device = self.framework_specific_info["device"]
        self.backend = self.framework_specific_info["backend"]
        self.recipes = deep_get(self.framework_specific_info, "recipes", {})

        self.pre_optimized_model = None
        self.pre_optimizer_handle = None
        self.bf16_ops = []
        self.fp32_ops = []
        self.query_handler = KerasQuery(local_config_file=os.path.join(os.path.dirname(__file__), "keras.yaml"))

        self.fp32_results = []
        self.fp32_preds_as_label = False
        self.callbacks = []

        self.conv_format = {}
        self.fold_conv = []
        self.keras3 = True if version1_gte_version2(tf.__version__, "2.16.1") else False
        if not os.path.exists(DEFAULT_WORKSPACE):
            os.makedirs(DEFAULT_WORKSPACE)
        self.tmp_dir = (DEFAULT_WORKSPACE + "tmp_model.keras") if self.keras3 else (DEFAULT_WORKSPACE + "tmp_model")

    def _set_weights(self, qmodel, layer_weights):
        """Set fp32 weights to qmodel."""
        for qlayer in qmodel.layers:
            if qlayer.get_weights():
                if qlayer.name in layer_weights:
                    qlayer.set_weights(layer_weights[qlayer.name])
                else:  # pragma: no cover
                    hit_layer = False
                    for sub_layer in qlayer.submodules:
                        if sub_layer.name in layer_weights:
                            qlayer.set_weights(layer_weights[sub_layer.name])
                            hit_layer = True
                            break
                    if not hit_layer:
                        raise ValueError("Can not match the module weights....")
        return qmodel

    def _check_quantize_format(self, model):
        """The function that checks format for conv ops."""
        input_layer_dict = {}
        layer_name_mapping = {}

        for layer in model.layers:
            layer_name_mapping[layer.name] = layer
            for node in layer._outbound_nodes:
                layer_name = node.operation.name if self.keras3 else node.outbound_layer.name
                if layer_name not in input_layer_dict:
                    input_layer_dict[layer_name] = [layer.name]
                else:
                    input_layer_dict[layer_name].append(layer.name)

        for layer in model.layers:
            if layer.__class__.__name__ in self.supported_op:
                self.conv_format[layer.name] = "s8"
                input_layer_names = input_layer_dict[layer.name]
                for input_layer_name in input_layer_names:
                    check_layer = layer_name_mapping[input_layer_name]
                    if check_layer.__class__.__name__ == "Activation" and check_layer.activation.__name__ in ["relu"]:
                        self.conv_format[layer.name] = "u8"
                        break

    def _fuse_bn_keras3(self, fuse_conv_bn, fp32_layers):  # pragma: no cover
        fuse_layers = []
        fused_bn_name = ""
        for idx, layer in enumerate(fp32_layers):
            if hasattr(layer, "_outbound_nodes"):
                if layer.name == fused_bn_name:
                    continue

                if layer.name in self.conv_weights.keys():
                    new_outbound_nodes = []
                    conv_weight = self.conv_weights[layer.name]
                    for outbound_node in layer._outbound_nodes:
                        outbound_layer = outbound_node.operation
                        if outbound_layer.__class__.__name__ in ("BatchNormalization"):
                            fused_bn_name = outbound_layer.name
                            bn_weight = self.bn_weights[fused_bn_name]
                            self.layer_weights[layer.name] = fuse_conv_bn(
                                conv_weight, bn_weight, layer.__class__.__name__, outbound_layer.epsilon
                            )
                            self.fold_conv.append(layer.name)
                            for node in outbound_layer._outbound_nodes:
                                new_outbound_nodes.append(node)
                        else:
                            new_outbound_nodes.append(outbound_node)
                    layer._outbound_nodes.clear()
                    for node in new_outbound_nodes:
                        layer._outbound_nodes.append(node)

                fuse_layers.append(layer)
            else:
                if (
                    idx > 0
                    and layer.__class__.__name__ == "BatchNormalization"
                    and fp32_layers[idx - 1].__class__.__name__ == "Conv2D"
                ):
                    conv_name = fp32_layers[idx - 1].name
                    conv_weight = self.conv_weights[conv_name]
                    bn_weight = self.bn_weights[layer.name]
                    conv_type = fp32_layers[idx - 1].__class__.__name__

                    self.layer_weights[conv_name] = fuse_conv_bn(conv_weight, bn_weight, conv_type, layer.epsilon)
                    self.fold_conv.append(conv_name)
                else:
                    fuse_layers.append(layer)

        return fuse_layers

    def _fuse_bn_keras2(self, fuse_conv_bn, fp32_layers):  # pragma: no cover
        fuse_layers = []
        for idx, layer in enumerate(fp32_layers):
            if hasattr(layer, "_inbound_nodes"):
                if layer.__class__.__name__ in ("BatchNormalization"):
                    for bn_inbound_node in layer._inbound_nodes:
                        inbound_layer = bn_inbound_node.inbound_layers
                        if inbound_layer.name in self.conv_weights.keys():
                            conv_layer = inbound_layer
                            conv_weight = self.conv_weights[conv_layer.name]
                            bn_weight = self.bn_weights[layer.name]

                            self.layer_weights[conv_layer.name] = fuse_conv_bn(
                                conv_weight, bn_weight, conv_layer.__class__.__name__, layer.epsilon
                            )
                            self.fold_conv.append(conv_layer.name)
                        else:
                            fuse_layers.append(layer)
                elif len(layer._inbound_nodes):
                    new_bound_nodes = []
                    # OpLambda node will have different bound node
                    if layer.__class__.__name__ in ("TFOpLambda", "SlicingOpLambda"):
                        fuse_layers.append(layer)
                    else:
                        for bound_node in layer._inbound_nodes:
                            inbound_layers = bound_node.inbound_layers
                            if not isinstance(inbound_layers, list):
                                inbound_layers = [inbound_layers]
                            for inbound_layer in inbound_layers:
                                if inbound_layer in self.bn_weights.keys():
                                    for bn_inbound_node in inbound_layer._inbound_nodes:
                                        bn_inbound_layer = bn_inbound_node.inbound_layers
                                        if bn_inbound_layer.name in self.conv_weights.keys():
                                            new_bound_nodes.append(bn_inbound_node)
                                        else:
                                            if bound_node not in new_bound_nodes:
                                                new_bound_nodes.append(bound_node)
                                else:
                                    new_bound_nodes.append(bound_node)

                        layer._inbound_nodes.clear()
                        for bound_node in new_bound_nodes:
                            layer._inbound_nodes.append(bound_node)
                        fuse_layers.append(layer)
                else:
                    fuse_layers.append(layer)
            else:
                if (
                    idx > 0
                    and layer.__class__.__name__ == "BatchNormalization"
                    and fp32_layers[idx - 1].__class__.__name__ == "Conv2D"
                ):
                    conv_name = fp32_layers[idx - 1].name
                    conv_weight = self.conv_weights[conv_name]
                    bn_weight = self.bn_weights[layer.name]
                    conv_type = fp32_layers[idx - 1].__class__.__name__

                    self.layer_weights[conv_name] = fuse_conv_bn(conv_weight, bn_weight, conv_type, layer.epsilon)
                    self.fold_conv.append(conv_name)
                else:
                    fuse_layers.append(layer)

        return fuse_layers

    def _fuse_bn(self, model):  # pragma: no cover
        """Fusing Batch Normalization."""
        model.save(self.tmp_dir)
        fuse_bn_model = tf.keras.models.load_model(self.tmp_dir)
        fp32_layers = fuse_bn_model.layers

        def fuse_conv_bn(conv_weight, bn_weight, conv_type="Conv2D", eps=1.0e-5):
            assert conv_type in [
                "Conv2D",
                "DepthwiseConv2D",
                "SeparableConv2D",
            ], "only support Conv2D, DepthwiseConv2D, SeparableConv2D..."
            if len(bn_weight) > 3:
                if conv_type == "DepthwiseConv2D":
                    gamma = bn_weight[0].reshape(1, 1, bn_weight[0].shape[0], 1)
                    var = bn_weight[3].reshape(1, 1, bn_weight[3].shape[0], 1)
                else:
                    gamma = bn_weight[0].reshape(1, 1, 1, bn_weight[0].shape[0])
                    var = bn_weight[3].reshape(1, 1, 1, bn_weight[3].shape[0])
                beta = bn_weight[1]
                mean = bn_weight[2]
            else:
                gamma = 1.0
                beta = bn_weight[0]
                mean = bn_weight[1]
                if conv_type == "DepthwiseConv2D":
                    var = bn_weight[2].reshape(1, 1, bn_weight[2].shape[0], 1)
                else:
                    var = bn_weight[2].reshape(1, 1, 1, bn_weight[2].shape[0])

            if len(conv_weight) == 1:
                weight = conv_weight[0]
                bias = np.zeros_like(beta)
            elif len(conv_weight) == 2 and conv_type == "SeparableConv2D":
                depth_weight = conv_weight[0]
                weight = conv_weight[1]
                bias = np.zeros_like(beta)
            elif len(conv_weight) == 2 and conv_type != "SeparableConv2D":
                weight = conv_weight[0]
                bias = conv_weight[1]
            elif len(conv_weight) == 3:
                depth_weight = conv_weight[0]
                weight = conv_weight[1]
                bias = conv_weight[2]
            scale_value = gamma / np.sqrt(var + eps)
            weight = weight * scale_value
            bias = beta + (bias - mean) * scale_value.reshape(-1)
            bias = bias.reshape(-1)
            return [depth_weight, weight, bias] if conv_type == "SeparableConv2D" else [weight, bias]

        fuse_bn_function = self._fuse_bn_keras3 if self.keras3 else self._fuse_bn_keras2
        fused_layers = fuse_bn_function(fuse_conv_bn, fp32_layers)

        for idx, layer in enumerate(fused_layers):
            if (
                layer.__class__.__name__ in ("Conv2D", "DepthwiseConv2D", "SeparableConv2D")
                and layer.name in self.fold_conv
            ):
                conv_config = layer.get_config()
                conv_config["use_bias"] = True
                conv_layer = type(layer).from_config(conv_config)
                for node in layer._outbound_nodes:
                    conv_layer._outbound_nodes.append(node)
                fused_layers[idx] = conv_layer

        bn_surgery = KerasSurgery(model)
        bn_fused_model = bn_surgery.convert(fused_layers, self.conv_weights.keys())
        bn_fused_model = self._set_weights(bn_fused_model, self.layer_weights)

        bn_fused_model.save(self.tmp_dir)
        bn_fused_model = tf.keras.models.load_model(self.tmp_dir)

        return bn_fused_model

    @dump_elapsed_time("Pass quantize model")
    def quantize(self, quant_config, model, dataloader, iteration, calib_func=None):
        """Execute the quantize process on the specified model.

        Args:
            quant_config(dict): The user defined 'StaticQuantConfig' class.
            model (object): The model to do quantization.
            dataloader(object): The calibration dataloader used to load quantization dataset.
            iteration(int): The iteration of calibration.
            calib_func (optional): the function used for calibration, should be a substitution for calibration
            dataloader when the built-in calibration function of INC does not work for model inference.
        """
        assert calib_func is None, "The calibration function is not supported on Keras backend yet"
        self.query_fw_capability(model)
        converter = KerasConfigConverter(quant_config, iteration)
        tune_cfg = converter.parse_to_tune_cfg()
        self.tuning_cfg_to_fw(tune_cfg)

        logger.debug("Dump quantization configurations:")
        logger.debug(self.quantize_config)
        calib_sampling_size = tune_cfg.get("calib_sampling_size", 1)

        if hasattr(dataloader, "batch_size") and calib_sampling_size % dataloader.batch_size != 0:
            iter = self.quantize_config["calib_iteration"]
            logger.warning(
                "Please note that calibration sampling size {} "
                "isn't divisible exactly by batch size {}. "
                "So the real sampling size is {}.".format(
                    calib_sampling_size, dataloader.batch_size, dataloader.batch_size * iter
                )
            )

        from neural_compressor.tensorflow.keras.layers import layer_initializer_dict

        q_layer_dict = {}
        for layer in self.pre_optimized_model.layers:
            if layer.__class__.__name__ in self.supported_op and layer.name in self.quantize_config["op_wise_config"]:
                op_config = self.quantize_config["op_wise_config"][layer.name]
                granularity = "per_channel" if op_config[0] else "per_tensor"
                q_layer_class = "Q" + layer.__class__.__name__
                q_config = {"T": self.conv_format[layer.name], "granularity": granularity}
                q_layer = layer_initializer_dict[q_layer_class](layer, q_config)
                q_layer_dict[layer.name] = q_layer

        calib_surgery = KerasSurgery(self.pre_optimized_model)
        calibration_model = calib_surgery.convert(q_layer_dict=q_layer_dict)
        calibration_model = self._set_weights(calibration_model, self.layer_weights)

        quantized_model = self._calibrate(
            calibration_model,
            dataloader,
            self.quantize_config["calib_iteration"],
        )

        return quantized_model

    def _calibrate(self, model, dataloader=None, calib_interation=None):
        """Apply calibration.

        Args:
            model(tf.keras.Model): The model inserted with FakeQuant layers for calibration.
            dataloader(object): The calibration dataloader used to load quantization dataset.
            calib_interation(int): The iteration of calibration.
        """
        # run eagerly to fetch the numpy min/max
        results = {}
        model.compile(run_eagerly=True)
        for idx, (inputs, _) in enumerate(dataloader):
            _ = model.predict_on_batch(inputs)
            for layer in model.layers:
                if (
                    layer.__class__.__name__[1:] in self.supported_op
                    and layer.name in self.quantize_config["op_wise_config"]
                ):
                    min_value = layer.act_min_value.numpy()
                    max_value = layer.act_max_value.numpy()

                    if layer.name not in results:
                        results[layer.name] = {"min": [min_value], "max": [max_value]}
                    else:
                        results[layer.name]["min"].append(min_value)
                        results[layer.name]["max"].append(max_value)
            if idx + 1 == calib_interation:
                break

        for idx, layer in enumerate(model.layers):
            if (
                layer.__class__.__name__[1:] in self.supported_op
                and layer.name in self.quantize_config["op_wise_config"]
            ):
                layer.act_min_value = np.min(results[layer.name]["min"])
                layer.act_max_value = np.max(results[layer.name]["max"])
                layer.quant_status = "quantize"
                # for layers that have weights
                if layer.name in self.layer_weights:
                    kernel = self.layer_weights[layer.name][0]
                    dim = list(range(0, kernel.ndim))
                    t_dim = [dim.pop(-1)]
                    t_dim.extend(dim)
                    channel_size = kernel.shape[-1]
                    kernel_channel = kernel.transpose(t_dim).reshape(channel_size, -1)

                    layer.weight_min_value = np.min(kernel_channel, axis=1).tolist()
                    layer.weight_max_value = np.max(kernel_channel, axis=1).tolist()
                else:
                    # default value, but never expected to be used
                    # cause no kernel weights for this layer
                    layer.weight_min_value = [-10000]
                    layer.weight_max_value = [10000]

        model.save(self.tmp_dir)
        quantized_model = tf.keras.models.load_model(self.tmp_dir)

        return quantized_model

    def query_fw_capability(self, model):
        """The function is used to return framework tuning capability.

        Args:
            model (object): The model to query quantization tuning capability.
        """
        if not isinstance(model, tf.keras.Model):
            model = model.model
        fp32_config = {"weight": {"dtype": "fp32"}, "activation": {"dtype": "fp32"}}
        bf16_config = {"weight": {"dtype": "bf16"}, "activation": {"dtype": "bf16"}}
        int8_type = self.query_handler.get_op_types_by_precision(precision="int8")
        op_capability = self.query_handler.get_quantization_capability()
        conv_config = copy.deepcopy(op_capability["int8"]["Conv2D"])
        conv_config = copy.deepcopy(op_capability["int8"]["SeparableConv2D"])
        conv_config = copy.deepcopy(op_capability["int8"]["DepthwiseConv2D"])
        dense_config = copy.deepcopy(op_capability["int8"]["Dense"])
        maxpool_config = copy.deepcopy(op_capability["int8"]["MaxPooling2D"])
        avgpool_config = copy.deepcopy(op_capability["int8"]["AveragePooling2D"])
        other_config = copy.deepcopy(op_capability["int8"]["default"])

        # # get fp32 layer weights
        self.fp32_model = model
        self.conv_weights = {}
        self.bn_weights = {}
        self.layer_weights = {}
        for layer in self.fp32_model.layers:
            if layer.get_weights():
                if (
                    isinstance(layer, tf.keras.layers.Conv2D)
                    or isinstance(layer, tf.keras.layers.DepthwiseConv2D)
                    or isinstance(layer, tf.keras.layers.SeparableConv2D)
                ):
                    self.conv_weights[layer.name] = copy.deepcopy(layer.get_weights())
                elif isinstance(layer, tf.keras.layers.BatchNormalization):
                    self.bn_weights[layer.name] = copy.deepcopy(layer.get_weights())
                self.layer_weights[layer.name] = copy.deepcopy(layer.get_weights())

        self._check_quantize_format(self.fp32_model)
        self.pre_optimized_model = self._fuse_bn(self.fp32_model)

        quantizable_layer_details = OrderedDict()
        for layer in self.fp32_model.layers:
            layer_class = layer.__class__.__name__
            if layer_class == "Conv2D":
                quantizable_layer_details[(layer.name, layer_class)] = [conv_config, bf16_config, fp32_config]
            elif layer_class == "Dense":
                quantizable_layer_details[(layer.name, layer_class)] = [dense_config, bf16_config, fp32_config]
            elif layer_class in {"AveragePooling2D", "AvgPool2D"}:
                quantizable_layer_details[(layer.name, layer_class)] = [avgpool_config, bf16_config, fp32_config]
            elif layer_class in {"MaxPooling2D", "MaxPool2D"}:
                quantizable_layer_details[(layer.name, layer_class)] = [maxpool_config, bf16_config, fp32_config]
            else:
                quantizable_layer_details[(layer.name, layer_class)] = [bf16_config, fp32_config]

        capability = {
            "opwise": copy.deepcopy(quantizable_layer_details),
            "optypewise": self.get_optype_wise_ability(quantizable_layer_details),
        }
        logger.debug("Dump framework quantization capability:")
        logger.debug(capability)

        return capability

    def get_optype_wise_ability(self, quantizable_op_details):
        """Get the op type wise capability by generating the union value of each op type.

        Returns:
            [string dict]: the key is op type while the value is the
                           detail configurations of activation and weight for this op type.
        """
        res = OrderedDict()
        for op in quantizable_op_details:
            if op[1] not in res:
                res[op[1]] = {"activation": quantizable_op_details[op][0]["activation"]}
                if "weight" in quantizable_op_details[op][0]:
                    res[op[1]]["weight"] = quantizable_op_details[op][0]["weight"]
        return res

    def tuning_cfg_to_fw(self, tuning_cfg):
        """Parse tune_config and set framework variables.

        Args:
            tuning_cfg (dict): The dict of tuning config.
        """
        self.quantize_config["calib_iteration"] = tuning_cfg["calib_iteration"]
        self.quantize_config["device"] = self.device
        self.quantize_config["advance"] = deep_get(tuning_cfg, "advance")
        fp32_ops = []
        bf16_ops = []
        bf16_type = set(self.query_handler.get_op_types_by_precision(precision="bf16"))
        dispatched_op_names = [j[0] for j in tuning_cfg["op"]]
        invalid_op_names = [i for i in self.quantize_config["op_wise_config"] if i not in dispatched_op_names]

        for op_name in invalid_op_names:
            self.quantize_config["op_wise_config"].pop(op_name)

        for each_op_info in tuning_cfg["op"]:
            op_name = each_op_info[0]

            if tuning_cfg["op"][each_op_info]["activation"]["dtype"] == "bf16":  # pragma: no cover
                if each_op_info[1] in bf16_type:
                    bf16_ops.append(op_name)
                continue

            if tuning_cfg["op"][each_op_info]["activation"]["dtype"] == "fp32":
                if op_name in self.quantize_config["op_wise_config"]:
                    self.quantize_config["op_wise_config"].pop(op_name)
                    fp32_ops.append(op_name)
                continue

            is_perchannel = False
            bit = None
            if "weight" in tuning_cfg["op"][each_op_info]:
                is_perchannel = tuning_cfg["op"][each_op_info]["weight"]["granularity"] == "per_channel"
                # bit = tuning_cfg['op'][each_op_info]['weight']['bit']
            weight_bit = bit if bit else 7.0
            algorithm = tuning_cfg["op"][each_op_info]["activation"]["algorithm"]
            is_asymmetric = False
            if "activation" in tuning_cfg["op"][each_op_info]:
                is_asymmetric = tuning_cfg["op"][each_op_info]["activation"]["scheme"] == "asym"
            self.quantize_config["op_wise_config"][op_name] = (is_perchannel, algorithm, is_asymmetric, weight_bit)
        self.bf16_ops = bf16_ops
        if self.bf16_ops:
            self.bf16_ops.pop(-1)
        self.fp32_ops = fp32_ops


class KerasQuery:
    """Class that queries configs from yaml settings."""

    def __init__(self, local_config_file=None):
        """Initialize KerasQuery."""
        self.version = tf.version.VERSION
        self.cfg = local_config_file
        self.cur_config = None
        self._one_shot_query()

    def _one_shot_query(self):
        """Query cur configs in one shot."""
        with open(self.cfg) as f:
            content = yaml.safe_load(f)
            try:
                self.cur_config = self._get_specified_version_cfg(content)
            except Exception as e:
                logger.info("Fail to parse {} due to {}.".format(self.cfg, str(e)))
                self.cur_config = None
                raise ValueError(
                    "Please check if the format of {} follows Neural Compressor yaml schema.".format(self.cfg)
                )

    def _get_specified_version_cfg(self, data):
        """Get the configuration for the current runtime.

        If there's no matched configuration in the input yaml, we'll
        use the `default` field of yaml.
        Args:
            data (Yaml content): input yaml file.

        Returns:
            [dictionary]: the content for specific version.
        """
        default_config = None
        for sub_data in data:
            if sub_data["version"]["name"] == self.version:
                return sub_data

            if sub_data["version"]["name"] == "default":
                default_config = sub_data

        return default_config

    def get_quantization_capability(self):
        """Get the supported op types' quantization capability.

        Returns:
            [dictionary list]: A list composed of dictionary which key is precision
            and value is a dict that describes all op types' quantization capability.
        """
        return self.cur_config["capabilities"]

    def get_op_types_by_precision(self, precision):
        """Get op types per precision.

        Args:
            precision (string): precision name

        Returns:
            [string list]: A list composed of op type.
        """
        assert precision in list(self.cur_config["ops"].keys())
        return self.cur_config["ops"][precision]


class KerasConfigConverter:
    """Convert `StaticQuantConfig` to the format used by static quant algo."""

    support_int8_weight = {"Dense", "Conv2D", "DepthwiseConv2D", "SeparableConv2D"}

    def __init__(self, quant_config: StaticQuantConfig, calib_iteration: int):
        """Init parser for keras static quant config.

        Args:
            quant_config: the keras static quant config.
            calib_iteration: the iteration of calibration.
        """
        self.quant_config = quant_config
        self.calib_iteration = calib_iteration

    def update_config(self, quant_config, op_key):
        """Update op-wise config.

        Args:
            quant_config: the keras static quant config.
            op_key: a tuple such as (layer type, layer name).
        """
        op_value = {"activation": {}}
        op_value["activation"].update(
            {
                "dtype": quant_config.act_dtype,
                "quant_mode": "static",
                "scheme": ("sym" if quant_config.act_sym else "asym"),
                "granularity": quant_config.act_granularity,
                "algorithm": "minmax",
            }
        )
        if op_key[1] not in self.support_int8_weight:
            return op_value

        op_value["weight"] = {
            "dtype": quant_config.weight_dtype,
            "scheme": "sym" if quant_config.weight_sym else "asym",
            "granularity": quant_config.weight_granularity,
            "algorithm": "minmax",
        }
        return op_value

    def parse_to_tune_cfg(self) -> Dict:
        """The function that parses StaticQuantConfig to keras tuning config."""
        tune_cfg = {"op": OrderedDict()}
        for op_key, config in self.quant_config.items():
            op_value = self.update_config(config, op_key)
            tune_cfg["op"].update({op_key: op_value})
            tune_cfg["calib_iteration"] = self.calib_iteration

        return tune_cfg


class KerasSurgery:
    """The class that inserts FakeQuant or QDQ layers before the target layers."""

    def __init__(self, model):
        """Init the KerasSurgery class.

        Args:
            model: the model to be modified.
        """
        import shutil

        self.model_outputs = []
        self.keras3 = True if version1_gte_version2(tf.__version__, "2.16.1") else False
        self.tmp_dir = (DEFAULT_WORKSPACE + "tmp_model.keras") if self.keras3 else (DEFAULT_WORKSPACE + "tmp_model")
        model.save(self.tmp_dir)
        self.model = tf.keras.models.load_model(self.tmp_dir)
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

    def _parse_inputs(self, BN_fused_layers=None, conv_names=None):
        """Create a input_layer_dict from model.

        Args:
            BN_fused_layers: The layers in which BN layers have been fused.
            conv_names: The name list of conv layers where BNs are fused.

        Returns:
            input_layer_dict: The dict that mapping for layer names to their input layer names.
        """
        input_layer_dict = {}
        layers = BN_fused_layers if BN_fused_layers else self.model.layers
        for layer in layers:
            for node in layer._outbound_nodes:
                out_layer = node.operation if self.keras3 else node.outbound_layer
                out_layer_names = [out_layer.name]
                if conv_names and out_layer.__class__.__name__ in ("BatchNormalization") and layer.name in conv_names:
                    out_layer_names = (
                        [node.operation.name for node in out_layer._outbound_nodes]
                        if self.keras3
                        else [node.outbound_layer.name for node in out_layer._outbound_nodes]
                    )

                for out_layer_name in out_layer_names:
                    if out_layer_name not in input_layer_dict:
                        input_layer_dict[out_layer_name] = [layer.name]
                    else:
                        input_layer_dict[out_layer_name].append(layer.name)

        try:
            model_input = self.model.input
        except ValueError:  # pragma: no cover
            model_input = self.model.inputs[0]

        return input_layer_dict, model_input

    def convert(self, BN_fused_layers=None, conv_names=None, q_layer_dict=None):
        """Generate optimized model by fusing BN layers or replacing Keras layers to custom quantized layers.

        Args:
            BN_fused_layers: The layers in which BN layers have been fused.
            conv_names: The name list of conv layers where BNs are fused.
            q_layer_dict: The dict mapping from keras layers to custom quantized layers.
        """
        input_layer_dict, model_input = self._parse_inputs(BN_fused_layers, conv_names)
        output_tensor_dict = {"keras.Input": model_input}
        layers = BN_fused_layers if BN_fused_layers else self.model.layers

        for idx, layer in enumerate(layers):
            if layer.__class__.__name__ == "InputLayer":
                output_tensor_dict[layer.name] = output_tensor_dict["keras.Input"]
                continue

            input_tensors = (
                output_tensor_dict["keras.Input"]
                if idx == 0
                else [output_tensor_dict[input_layer] for input_layer in input_layer_dict[layer.name]]
            )

            while isinstance(input_tensors, list) and len(input_tensors) == 1:
                input_tensors = input_tensors[0]

            if self.keras3:
                layer._inbound_nodes.clear()

            cur_layer = q_layer_dict[layer.name] if q_layer_dict and layer.name in q_layer_dict else layer
            x = cur_layer(input_tensors)
            output_tensor_dict[layer.name] = x

            if not isinstance(self.model, tf.keras.models.Sequential) and layer.name in self.model.output_names:
                self.model_outputs.append(x)

        if not self.model_outputs:
            self.model_outputs.append(x)

        return tf.keras.models.Model(inputs=self.model.inputs, outputs=self.model_outputs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse small ops to Gelu Graph Rewriter."""

import tensorflow as tf
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.graph_base import GraphRewriterBase
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper


class FuseGeluOptimizer(GraphRewriterBase):  # pragma: no cover
    """Fuse Sqrt + RealDiv + Erf + AddV2 + Mul + Mul into Gelu op."""

    def do_transformation(self):
        """Execute the fusion from small ops to Gelu."""
        if tf.version.VERSION not in ("1.15.0-up2", "1.15.0-up3"):
            return self.model

        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()
        # Below code is relative to expression on
        # https://github.com/IntelAI/models/blob/master/models/language_modeling/tensorflow/
        # bert_large/inference/generic_ops.py#L105
        target_nodes = cur_graph.query_fusion_pattern_nodes(
            [["Pow"], ["Mul"], ["AddV2"], ["Mul"], ["Tanh"], ["AddV2"], ["Mul"], ["Mul"]]
        )

        if not target_nodes:
            target_nodes = cur_graph.query_fusion_pattern_nodes(
                [["Pow"], ["Mul"], ["AddV2"], ["Mul"], ["Tanh"], ["AddV2"], ["Mul"]]
            )

        for node_combination in target_nodes:
            match_node_length = len(node_combination)
            pow_node = graph_info[node_combination[0]].node
            mul_1_node = graph_info[node_combination[1]].node
            addv2_1_node = graph_info[node_combination[2]].node
            mul_2_node = graph_info[node_combination[3]].node
            tanh_node = graph_info[node_combination[4]].node
            addv2_2_node = graph_info[node_combination[5]].node
            if match_node_length == 8:
                mul_3_node = graph_info[node_combination[6]].node
            else:
                mul_3_node = graph_info[node_combination[7]].node

            gelu_input_name = None
            pow_const_node_name = None
            pow_value = None

            for i in pow_node.input:
                node_name = Helper.node_name_from_input(i)
                if graph_info[node_name].node.op != "Const":
                    gelu_input_name = i

                if graph_info[node_name].node.op == "Const":
                    pow_const_node_name = i
                    pow_value = graph_info[node_name].node.attr["value"].tensor.float_val[0]
                    break

            if pow_value != 3:
                continue
            mul_1_value = None
            mul_1_const_node_name = None
            for i in mul_1_node.input:
                i = Helper.node_name_from_input(i)
                if i != pow_node.name and graph_info[i].node.op == "Const":
                    mul_1_const_node_name = i
                    mul_1_value = graph_info[i].node.attr["value"].tensor.float_val[0]
                    break
            if mul_1_value != 0.044714998453855515:
                continue

            mul_2_value = None
            mul_2_const_node_name = None
            for i in mul_2_node.input:
                i = Helper.node_name_from_input(i)
                if i != addv2_1_node.name and graph_info[i].node.op == "Const":
                    mul_2_const_node_name = i
                    mul_2_value = graph_info[i].node.attr["value"].tensor.float_val[0]
                    break
            if mul_2_value != 0.7978845834732056:
                continue

            addv2_2_value = None
            addv2_2_const_node_name = None
            for i in addv2_2_node.input:
                i = Helper.node_name_from_input(i)
                if i != tanh_node.name and graph_info[i].node.op == "Const":
                    addv2_2_const_node_name = i
                    addv2_2_value = graph_info[i].node.attr["value"].tensor.float_val[0]
                    break
            if addv2_2_value != 1:
                continue

            rest_mul_node = None
            if match_node_length == 8:
                for i in mul_3_node.input:
                    i = Helper.node_name_from_input(i)
                    if i != addv2_2_node.name:
                        rest_mul_node = graph_info[i].node
                        break

                if not rest_mul_node or rest_mul_node.op != "Mul":
                    continue
            else:
                rest_mul_node = graph_info[node_combination[6]].node

            rest_mul_valid = False
            rest_mul_const_node_name = None
            for i in rest_mul_node.input:
                i = Helper.node_name_from_input(i)
                if graph_info[i].node.op == "Const" and graph_info[i].node.attr["value"].tensor.float_val[0] == 0.5:
                    rest_mul_const_node_name = i
                    rest_mul_valid = True
                    break

            if not rest_mul_valid:
                continue

            cur_graph.remove_node(pow_const_node_name)
            cur_graph.remove_node(pow_node.name)
            cur_graph.remove_node(mul_1_node.name)
            cur_graph.remove_node(mul_1_const_node_name)
            cur_graph.remove_node(addv2_1_node.name)
            cur_graph.remove_node(mul_2_node.name)
            cur_graph.remove_node(mul_2_const_node_name)
            cur_graph.remove_node(tanh_node.name)
            cur_graph.remove_node(addv2_2_node.name)
            cur_graph.remove_node(addv2_2_const_node_name)
            cur_graph.remove_node(rest_mul_node.name)
            cur_graph.remove_node(rest_mul_const_node_name)

            original_last = graph_info[mul_3_node.name].outputs
            cur_graph.remove_node(mul_3_node.name)
            gelu_node = Helper.create_node("Gelu", mul_3_node.name, [gelu_input_name])
            Helper.set_attr_bool(gelu_node, "approximate", True)
            Helper.set_attr_dtype(gelu_node, "T", dtypes.float32)

            cur_graph.add_node(gelu_node, gelu_input_name, original_last)

        target_nodes = cur_graph.query_fusion_pattern_nodes(
            [["Sqrt"], ["RealDiv"], ["Erf"], ["AddV2"], ["Mul"], ["Mul"]]
        )

        for node_combination in target_nodes:
            sqrt_node = graph_info[node_combination[0]].node
            realdiv_node = graph_info[node_combination[1]].node
            erf_node = graph_info[node_combination[2]].node
            addv2_node = graph_info[node_combination[3]].node
            mul1_node = graph_info[node_combination[4]].node
            mul2_node = graph_info[node_combination[5]].node

            sqrt_input_name = Helper.node_name_from_input(sqrt_node.input[0])
            sqrt_value = graph_info[sqrt_input_name].node.attr["value"].tensor.float_val[0]

            if sqrt_value != 2:
                continue

            addv2_value = None
            mul1_value = None
            gelu_input_name = None

            for i in realdiv_node.input:
                i = Helper.node_name_from_input(i)
                if i != sqrt_node.name:
                    gelu_input_name = i
                    break

            addv2_const_name = None
            for i in addv2_node.input:
                i = Helper.node_name_from_input(i)
                if i != erf_node.name:
                    addv2_value = graph_info[i].node.attr["value"].tensor.float_val[0]
                    addv2_const_name = i
                    break

            if addv2_value != 1:
                continue

            mul1_const_name = None
            for i in mul1_node.input:
                i = Helper.node_name_from_input(i)
                if i != addv2_node.name:
                    mul1_value = graph_info[i].node.attr["value"].tensor.float_val[0]
                    mul1_const_name = i
                    break

            if mul1_value != 0.5:
                continue

            cur_graph.remove_node(sqrt_node.input[0])
            cur_graph.remove_node(sqrt_node.name)
            cur_graph.remove_node(realdiv_node.name)
            cur_graph.remove_node(erf_node.name)
            cur_graph.remove_node(addv2_node.name)
            cur_graph.remove_node(mul1_node.name)
            cur_graph.remove_node(addv2_const_name)
            cur_graph.remove_node(sqrt_input_name)
            cur_graph.remove_node(mul1_const_name)

            original_last = graph_info[mul2_node.name].outputs
            cur_graph.remove_node(mul2_node.name)
            gelu_node = Helper.create_node("Gelu", mul2_node.name, [gelu_input_name])
            Helper.set_attr_bool(gelu_node, "approximate", False)
            Helper.set_attr_dtype(gelu_node, "T", dtypes.float32)

            cur_graph.add_node(gelu_node, gelu_input_name, original_last)

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert Layout Graph Rewriter."""

import tensorflow as tf
from tensorflow.core.protobuf import config_pb2, meta_graph_pb2, rewriter_config_pb2
from tensorflow.python.grappler import tf_optimizer
from tensorflow.python.training import saver as saver_lib

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.graph_base import GraphRewriterBase
from neural_compressor.tensorflow.utils import dump_elapsed_time, version1_gt_version2


class ConvertLayoutOptimizer(GraphRewriterBase):
    """The layout conversion optimizer, convert NCHW to NHWC format.

    It is executed only when NCHW node exists and tensorflow version is 2.4.0 and above.

    Args: model: input graph_def
          outputs: output name list

    Return: converted graph_def
    """

    def __init__(self, model, outputs):
        """Initialization."""
        super().__init__(model)
        self.outputs = outputs

    @dump_elapsed_time("ConvertLayoutOptimizer")
    def do_transformation(self):
        """Execute converting layout."""
        convert = False
        for node in self.model.node:
            if "Conv" in node.op and "data_format" in node.attr and node.attr["data_format"].s == b"NCHW":
                convert = True
                break
        if convert and version1_gt_version2(tf.version.VERSION, "2.3.0"):
            g = tf.Graph()
            with g.as_default():  # pylint: disable=not-context-manager
                g = tf.compat.v1.import_graph_def(self.model, name="")
                meta_graph = saver_lib.export_meta_graph(graph_def=self.model, graph=g, clear_devices=False)
                fetch_collection = meta_graph_pb2.CollectionDef()
                for fetch in self.outputs:
                    fetch_collection.node_list.value.append(fetch)  # pylint: disable=no-member
                meta_graph.collection_def["train_op"].CopyFrom(  # pylint: disable=no-member
                    fetch_collection
                )  # pylint: disable=no-member

            config = config_pb2.ConfigProto()
            convert = rewriter_config_pb2.RewriterConfig.NCHW_TO_NHWC  # pylint: disable=no-member
            config.graph_options.rewrite_options.CopyFrom(  # pylint: disable=no-member
                rewriter_config_pb2.RewriterConfig(
                    disable_model_pruning=True,
                    constant_folding=rewriter_config_pb2.RewriterConfig.OFF,
                    dependency_optimization=rewriter_config_pb2.RewriterConfig.OFF,
                    memory_optimization=rewriter_config_pb2.RewriterConfig.NO_MEM_OPT,
                    arithmetic_optimization=rewriter_config_pb2.RewriterConfig.OFF,
                    shape_optimization=rewriter_config_pb2.RewriterConfig.OFF,
                    loop_optimization=rewriter_config_pb2.RewriterConfig.OFF,
                    function_optimization=rewriter_config_pb2.RewriterConfig.OFF,
                    remapping=rewriter_config_pb2.RewriterConfig.OFF,
                    implementation_selector=rewriter_config_pb2.RewriterConfig.OFF,
                    cpu_layout_conversion=convert,
                )
            )

            optimized_graph = tf_optimizer.OptimizeGraph(config, meta_graph)
            return optimized_graph

        return self.model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantize Graph Common Utils Herlper Class."""

import re

import numpy as np
from tensorflow.core.framework import attr_value_pb2, graph_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util


class QuantizeGraphHelper:
    """This class contains several staticmethod functions."""

    node_name_cache = {}
    node_name_port_cache = {}

    def __init__(self):
        """Initialization."""
        pass

    def _recursive_graph_sorting(self, node_name):
        """Recursive sort the graph."""
        if node_name in self.op_list or not self.node_name_mapping[node_name].input:
            return

        for input_name in self.node_name_mapping[node_name].input:
            if input_name not in self.node_name_mapping:
                continue
            else:
                self._recursive_graph_sorting((input_name))

        if node_name not in self.op_list:
            self.op_list.append(node_name)

        return

    def _get_op_list(self, output_node_names):
        """Get op list by recursive sorting the graph."""
        for output_name in output_node_names:
            self._recursive_graph_sorting(output_name)

    def get_sorted_graph(self, input_graph, input_node_names, output_node_names):
        """Return a sorted graphdef object.

        Sometimes the input graphdef was composed of the random nodedef objects,
        we reorder the graph to make the parsing easier.

        Args:
            input_graph (graphdef]): the input graphdef object
            input_node_names (string list): the input node names
            output_node_names (string list): the output node names

        Returns:
            [type]: [description]
        """
        self.node_name_mapping = {}
        self.op_list = [input_node_name for input_node_name in input_node_names]
        for node in input_graph.node:
            self.node_name_mapping[node.name] = node
        self._get_op_list(output_node_names)

        all_ops = [i for i in list(self.node_name_mapping.keys()) if i not in self.op_list]
        self.op_list.extend(sorted(set(all_ops), key=all_ops.index))

        self.out_graph_def = graph_pb2.GraphDef()
        for i in self.op_list:
            new_node = node_def_pb2.NodeDef()
            new_node.CopyFrom(self.node_name_mapping[i])
            self.out_graph_def.node.extend([new_node])

        return self.out_graph_def

    @staticmethod
    def split_shared_inputs(input_graph_def):
        """Split shared inputs(like weights and bias) of the graph.

        :param in_graph: input graph file.
        :return: path to output graph file.
        """
        node_map = {}
        for node in input_graph_def.node:
            if node.name not in node_map:
                node_map[node.name] = node

        output_graph_def = graph_pb2.GraphDef()
        is_shared_input = False
        # map of: input_name - op_name
        input_map = {}
        for node_name in node_map.keys():
            node = node_map[node_name]
            for input_idx, input_node_name in enumerate(node.input):
                if node_map[QuantizeGraphHelper.node_name_from_input(input_node_name)].op == "Const":
                    # is shared and current node is not the first one
                    # sharing the input
                    if input_node_name in input_map.keys():
                        is_shared_input = True
                        input_map[input_node_name].append(node.name)
                        new_input_node = node_def_pb2.NodeDef()
                        new_input_node.CopyFrom(node_map[input_node_name])
                        new_input_node.name = input_node_name + "_" + str(len(input_map[input_node_name]))
                        node.input[input_idx] = new_input_node.name
                        output_graph_def.node.extend([new_input_node])
                    else:
                        input_map[input_node_name] = [node.name]
            output_graph_def.node.extend([node])

        return output_graph_def if is_shared_input else input_graph_def

    @staticmethod
    def remove_training_nodes(input_graph, protected_nodes=[], types_to_splice=["Identity", "CheckNumerics"]):
        """Prunes out nodes that aren't needed for inference.

        Args:
            input_graph: Model to analyze and prune.
            types_to_splice: An optional list of types of nodes to be removed
            unconditionally.

        Returns:
            A optimized graphdef object.
        """
        input_nodes = input_graph.node

        control_input_names = set()
        node_names_with_control_input = set()
        for node in input_nodes:
            for node_input in node.input:
                if "^" in node_input:
                    control_input_names.add(node_input.replace("^", ""))
                    node_names_with_control_input.add(node.name)

        names_to_splice = {}
        for node in input_nodes:
            if node.op in types_to_splice:
                # We don't want to remove nodes that have control edge inputs, because
                # they might be involved in subtle dependency issues that removing them
                # will jeopardize.
                if node.name not in node_names_with_control_input:
                    names_to_splice[node.name] = node.input[0]

        # We also don't want to remove nodes which are used as control edge inputs.
        names_to_splice = {name: value for name, value in names_to_splice.items() if name not in control_input_names}

        nodes_after_splicing = []

        for node in input_nodes:
            if node.name in names_to_splice and node.name not in protected_nodes:
                continue

            if node.name in protected_nodes and node.name in types_to_splice:
                nodes_after_splicing.append(node)
                continue

            new_node = node_def_pb2.NodeDef()
            new_node.CopyFrom(node)
            input_before_removal = node.input
            del new_node.input[:]
            for full_input_name in input_before_removal:
                input_name = re.sub(r"^\^", "", full_input_name)
                while input_name in names_to_splice:
                    full_input_name = names_to_splice[input_name]
                    input_name = re.sub(r"^\^", "", full_input_name)
                new_node.input.append(full_input_name)
            nodes_after_splicing.append(new_node)

        output_graph = graph_pb2.GraphDef()
        output_graph.node.extend(nodes_after_splicing)
        return output_graph

    @staticmethod
    def create_node(op, name, inputs):
        """Create a nodedef object.

        Args:
            op (string): op type
            name (string): op name
            inputs (string list): op's inputs name

        Returns:
            nodedef: the created nodedef object
        """
        new_node = node_def_pb2.NodeDef()
        new_node.op = op
        new_node.name = name
        for input_name in inputs:
            new_node.input.extend([input_name])
        return new_node

    @staticmethod
    def create_constant_node(name, value, dtype, shape=None, device="cpu"):
        """Create constant node.

        Args:
            name (string): op name
            value (np.array): input data
            dtype (datatype): data type of the input value
            shape (int list, optional): the value's shape. Defaults to None.
            device (str, optional): the device type, it may be the 'cpu' or 'gpu'.
                                    Defaults to 'cpu'.

        Returns:
            [type]: [description]
        """
        node = QuantizeGraphHelper.create_node("Const" if device == "cpu" else "HostConst", name, [])
        QuantizeGraphHelper.set_attr_dtype(node, "dtype", dtype)
        QuantizeGraphHelper.set_attr_tensor(node, "value", value, dtype, shape)
        return node

    @staticmethod
    def copy_attr(node, key, attr_value):
        """Copy the specified attr value to node.

        Args:
            node (nodedef): a nodedef object
            key (string): string name
            attr_value (any): the specified attribute value
        """
        node.attr[key].CopyFrom(attr_value)

    @staticmethod
    def set_attr_dtype(node, key, value):
        """Set the attribute data type."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(type=value.as_datatype_enum))

    @staticmethod
    def set_attr_tensor(node, key, value, dtype, shape=None):
        """Set the tensor value to specified attribute field.

        Args:
            node (nodedef): the target nodedef object
            key (string): attribute name
            value (np.array): the content
            dtype (dtypes): data type
            shape (int list, optional): the input tensor's shape. Defaults to None.
        """
        node.attr[key].CopyFrom(
            attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(value, dtype=dtype, shape=shape))
        )

    @staticmethod
    def set_attr_string_list(node, key, value):
        """Set the node's attr which data type is int list."""
        list_value = attr_value_pb2.AttrValue.ListValue(s=value)
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(list=list_value))

    @staticmethod
    def set_attr_type_list(node, key, value):
        """Set the node's attr which data type is int list."""
        list_value = attr_value_pb2.AttrValue.ListValue(type=value)
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(list=list_value))

    @staticmethod
    def set_attr_string(node, key, value):
        """Set the node's attr which data type is string."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(s=value))

    @staticmethod
    def set_attr_bool(node, key, value):
        """Set the node's attr which data type is bool."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(b=value))

    @staticmethod
    def set_attr_int(node, key, value):
        """Set the node's attr which data type is int."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(i=value))

    @staticmethod
    def set_attr_float(node, key, value):
        """Set the node's attr which data type is float."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(f=value))

    @staticmethod
    def node_name_from_input(node_name):
        """Static method that get the valid node name from input name.

        Args:
            node_name (string): node name defined in the input field.

        Returns:
            string: node's name
        """
        if node_name not in QuantizeGraphHelper.node_name_cache:
            key = node_name
            if node_name.startswith("^"):
                node_name = node_name[1:]
            m = re.search(r"(.*):\d+$", node_name)
            if m:
                node_name = m.group(1)
            QuantizeGraphHelper.node_name_cache[key] = node_name
            return node_name

        return QuantizeGraphHelper.node_name_cache[node_name]

    @staticmethod
    def unique_node_name_from_input(node_name):
        """Get the node name from other node name's input field."""
        return node_name.replace(":", "__port__").replace("^", "__hat__")

    @staticmethod
    def ensure_tensor_name_has_port(node_name):
        """Makes sure that a tensor name has :0 if no explicit port exists."""
        if node_name not in QuantizeGraphHelper.node_name_port_cache:
            key = node_name
            m = re.search(r"(.*):\d+$", node_name)
            if not m:
                node_name = node_name + ":0"
            QuantizeGraphHelper.node_name_port_cache[key] = node_name
            return node_name

        return QuantizeGraphHelper.node_name_port_cache[node_name]

    @staticmethod
    def generate_quantized_weight_node(
        host_op_type, input_node, per_channel, weight_bit=7.0, device="cpu", enter_node=None
    ):
        """Generated the quantized weight node."""
        base_name = input_node.name + "_"
        qint8_const_name = base_name + "qint8_const"
        min_name = base_name + "min"
        max_name = base_name + "max"
        float_tensor = tensor_util.MakeNdarray(input_node.attr["value"].tensor)
        epsilon = 1e-4  # Needs to be set empirically if accuracy is not satisfactory
        range_coefficent = 127 / (2**weight_bit - 1)
        if host_op_type in (
            "Conv2D",
            "MatMul",
            "Conv3D",
            "BatchMatMulV2",
            "Conv2DBackpropInput",
            "Conv3DBackpropInputV2",
        ):
            if per_channel:
                if host_op_type in ("Conv3D", "Conv3DBackpropInputV2"):
                    ranges = np.abs(float_tensor).max(axis=(0, 1, 2, 3))
                elif host_op_type in ("Conv2D", "Conv2DBackpropInput"):
                    ranges = np.abs(float_tensor).max(axis=(0, 1, 2))
                elif host_op_type in ("MatMul"):
                    if "transpose_b" in input_node.attr and input_node.attr["transpose_b"].b:  # pragma: no cover
                        ranges = np.abs(float_tensor).max(axis=(1))
                    else:
                        ranges = np.abs(float_tensor).max(axis=(0))
                else:
                    ranges = np.abs(float_tensor).max(axis=(0, 1, 2))

                ranges *= range_coefficent
                min_value = -ranges
                max_value = ranges
                # nudging min-max values outside epsilon radius around zero
                ranges[ranges < epsilon] = epsilon
                min_value[np.abs(min_value) < epsilon] = -epsilon
                max_value[np.abs(max_value) < epsilon] = epsilon
                if "transpose_b" in input_node.attr and input_node.attr["transpose_b"].b:  # pragma: no cover
                    # transpose for broadcasting
                    float_tensor = np.transpose(float_tensor, [1, 0])
                    qint8_tensor = (np.around(float_tensor * 127.0 / ranges)).astype(np.int8)
                    qint8_tensor = np.transpose(qint8_tensor, [1, 0])
                else:
                    qint8_tensor = (np.around(float_tensor * 127.0 / ranges)).astype(np.int8)
            else:
                min_value = np.min(float_tensor)
                max_value = np.max(float_tensor)
                min_value *= range_coefficent
                max_value *= range_coefficent
                min_value = min(min_value, 0.0)
                if min_value == max_value:
                    if abs(min_value) < 0.000001:
                        max_value = min_value + 1.0
                    elif min_value > 0:
                        max_value = 2 * min_value
                    else:
                        max_value = min_value / 2.0
                range_value = np.max(np.abs([min_value, max_value]))
                qint8_tensor = (np.around(float_tensor * 127.0 / range_value)).astype(np.int8)
                qint8_tensor = np.clip(qint8_tensor, -127, 127).astype(np.int8)
                min_value = -range_value
                max_value = range_value
        elif host_op_type == "DepthwiseConv2dNative":
            # get the max values based on dim 0 and 1 for depthwise conv
            # since, the output channel will be dim 2 * dim 3
            ranges = np.abs(float_tensor).max(axis=(0, 1))
            ranges = ranges.flatten()
            min_value = -ranges
            max_value = ranges
            # nudging min-max values outside epsilon radius around zero
            ranges[ranges < epsilon] = epsilon
            min_value[np.abs(min_value) < epsilon] = -epsilon
            max_value[np.abs(max_value) < epsilon] = epsilon
            # Since output channel will be 1 dim which is dim 2 * dim 3
            # When divide by range, qint8_tensor needs to be 3 dim
            # where, 3rd dim should be same dim of ranges
            a, b, c, d = float_tensor.shape
            qint8_tensor = (np.around(float_tensor.reshape(a, b, c * d) * 127.0 / ranges)).astype(np.int8)
            # get the shape back to 4 dim
            qint8_tensor = qint8_tensor.reshape(a, b, c, d)
        shape = tensor_util.TensorShapeProtoToList(input_node.attr["value"].tensor.tensor_shape)
        qint8_const_node = QuantizeGraphHelper.create_constant_node(
            qint8_const_name, qint8_tensor, dtypes.qint8, shape=shape
        )
        min_node = QuantizeGraphHelper.create_constant_node(min_name, min_value, dtypes.float32, device="cpu")

        max_node = QuantizeGraphHelper.create_constant_node(max_name, max_value, dtypes.float32, device="cpu")

        qint8_const_enter_node = None
        min_enter_node = None
        max_enter_node = None

        if enter_node:
            qint8_const_enter_node = QuantizeGraphHelper.create_node(
                "Enter", qint8_const_name + "_enter", [qint8_const_name]
            )
            QuantizeGraphHelper.set_attr_string(qint8_const_enter_node, "frame_name", enter_node.attr["frame_name"].s)
            QuantizeGraphHelper.set_attr_dtype(qint8_const_enter_node, "T", dtypes.qint8)
            QuantizeGraphHelper.set_attr_bool(qint8_const_enter_node, "is_constant", True)
            QuantizeGraphHelper.set_attr_int(
                qint8_const_enter_node, "parallel_iterations", enter_node.attr["parallel_iterations"].i
            )

            min_enter_node = QuantizeGraphHelper.create_node("Enter", min_name + "_enter", [min_name])
            QuantizeGraphHelper.set_attr_string(min_enter_node, "frame_name", enter_node.attr["frame_name"].s)
            QuantizeGraphHelper.set_attr_dtype(min_enter_node, "T", dtypes.float32)
            QuantizeGraphHelper.set_attr_bool(min_enter_node, "is_constant", True)
            QuantizeGraphHelper.set_attr_int(
                min_enter_node, "parallel_iterations", enter_node.attr["parallel_iterations"].i
            )

            max_enter_node = QuantizeGraphHelper.create_node("Enter", max_name + "_enter", [max_name])
            QuantizeGraphHelper.set_attr_string(max_enter_node, "frame_name", enter_node.attr["frame_name"].s)
            QuantizeGraphHelper.set_attr_dtype(max_enter_node, "T", dtypes.float32)
            QuantizeGraphHelper.set_attr_bool(max_enter_node, "is_constant", True)
            QuantizeGraphHelper.set_attr_int(
                max_enter_node, "parallel_iterations", enter_node.attr["parallel_iterations"].i
            )

        return qint8_const_node, min_node, max_node, qint8_const_enter_node, min_enter_node, max_enter_node



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow scaling model weights and activations for Smooth Quantization."""

import logging

import numpy as np
import tensorflow as tf
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer

logger = logging.getLogger("neural_compressor")


class SmoothQuantScaler:
    """A class for scaling model weights using Smooth Quantization method.

    Args:
        model: Tensorflow model to be scaled
        dataloader: Tensorflow dataloader for the dataset
        alpha: float, the scaling factor
        scales_per_op: bool, each op will have an individual scale or
                       ops with the same input will share a scale
    """

    def __init__(self, model, dataloader, alpha, scales_per_op):
        """Initialization."""
        self.model = model
        self.dataloader = dataloader
        self.alpha = alpha
        self.scales_per_op = scales_per_op
        self.mul_list = []
        self.g_analyzer = GraphAnalyzer()
        self.g_analyzer.graph = self.model

    def _adjust_activation(self, scale, input_node_name, output_node_name, w_i):
        """Insert the Mul node after the activation before the weight node.

        Args:
            scale: smooth scale with the shape (ic,)
            input_node_name: the parent input node
            output_node_name: the concrete output weight node name
            w_i: distinguish between different output weight nodes on different branches when naming
        """
        from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper

        node_suffix = str(w_i)
        mul_const_node = Helper.create_constant_node(input_node_name + "/scale_mul" + node_suffix, scale, tf.float32)
        mul_node = Helper.create_node(
            "Mul",
            input_node_name + "_mul" + node_suffix,
            [input_node_name + "/scale_mul" + node_suffix, input_node_name],
        )
        Helper.set_attr_dtype(mul_node, "T", dtypes.float32)
        self.mul_list.append(mul_node.name)
        self.g_analyzer.add_node(mul_node, input_node_name, [output_node_name])
        self.g_analyzer.add_node(mul_const_node, None, [input_node_name + "_mul" + node_suffix])

    def _adjust_weight(self, scale, weight_node, original_weight):
        """In-place adjust weight by scale.

        Args:
            scale: smooth scale with the shape (ic,)
            weight_node: reference to the original const weight node
            original_weight: numpy value of the original const weight node
        """
        # scale: (ic,)
        original_shape = original_weight.shape
        if len(original_shape) == 4:  # (fh, hw, ic, oc)
            W = np.transpose(original_weight, [0, 1, 3, 2])  # move input channel to last dimension
            W *= scale
            W = np.transpose(W, [0, 1, 3, 2])  # move input channel back
            weight_node.attr["value"].tensor.CopyFrom(tensor_util.make_tensor_proto(W))
        elif len(original_shape) == 2:  # (ic, oc) if transpose_a == transpose_b == false
            W = np.transpose(original_weight, [1, 0])
            W *= scale
            W = np.transpose(W, [1, 0])
            weight_node.attr["value"].tensor.CopyFrom(tensor_util.make_tensor_proto(W))

    def transform(self, max_vals_per_channel, sq_weight_tensors, sq_weights_nodes, sq_weight_node_names):
        """Apply scaling to weights and activations based on the maximum values per channel.

        Args:
            max_vals_per_channel (dict): A dictionary containing the maximum values per channel for each input node.
            sq_weight_tensors (dict): A dictionary containing the name -> weight tensors mapping for each input node.
            sq_weights_nodes (dict): A dictionary containing the name -> constant nodes mapping for each input node.
            sq_weight_node_names (dict): A dictionary from weight node name to the its concrete output node name.

        Returns:
            tuple: A tuple containing the modified model and a list of the inserted multiplication nodes.
        """
        logger.info("Start scaling on model graph for Smooth Quantization.")
        if self.scales_per_op:
            # 1. obtain the smooth scale per op
            # 2. adjust weight
            # 3. adjust activation
            for idx, input_node_name in enumerate(max_vals_per_channel):
                A_max_per_in_channel = max_vals_per_channel[input_node_name]
                W_dict = sq_weight_tensors[input_node_name]
                # Use the const nodes before to get weight values
                W_const_node_dict = sq_weights_nodes[input_node_name]
                # Get the concrete weight node as the output of Mul insertion
                for w_i, W_name in enumerate(W_dict):
                    W = W_dict[W_name]
                    if len(W.shape) == 4:
                        # https://www.tensorflow.org/api_docs/python/tf/nn/conv2d
                        # weight: [filter_height, filter_width, in_channels, out_channels]
                        # activation: NHWC, also batch_shape + [in_height, in_width, in_channels]
                        tensor = np.abs(np.transpose(W, [0, 1, 3, 2]))
                        # reduce weight max to (in_channel, ), aligned with activation max
                        W_max_per_in_channel = np.max(np.reshape(tensor, (-1, tensor.shape[-1])), axis=0)
                    elif len(W.shape) == 2:  # matmul
                        # reduce weight max to (in_channel, ), aligned with activation max
                        tensor = np.abs(W)
                        W_max_per_in_channel = np.max(tensor, axis=1)
                    else:  # pragma: no cover
                        assert False, "not supported"
                    cur_const_node = W_const_node_dict[W_name]
                    try:
                        scale = np.power(A_max_per_in_channel, self.alpha) / np.power(
                            W_max_per_in_channel, (1 - self.alpha)
                        )
                    except ValueError as e:  # pragma: no cover
                        logger.info(e)
                        logger.info("Skip smoothing the node: {}".format(cur_const_node.name))
                        continue
                    # clip the scales that are too small
                    scale = np.clip(scale, a_min=1e-5, a_max=1e8)
                    # skip smoothing the op where scale has elements that less than 1
                    # if np.any(scale < 1):
                    #     logger.info("skip smooth quant: {}".format(input_node_name))
                    #     continue
                    self._adjust_weight(scale, cur_const_node, W)
                    self._adjust_activation(1 / scale, input_node_name, sq_weight_node_names[cur_const_node.name], w_i)
        else:
            pass
        sq_graph_def = self.g_analyzer.dump_graph()
        sq_graph_def.library.CopyFrom(self.model.graph_def.library)
        self.model.graph_def = sq_graph_def
        return self.model, self.mul_list


class SmoothQuantScalerLLM(SmoothQuantScaler):
    """A class for scaling model weights for TF LLM models using Smooth Quantization method.

    Args:
        graph_def: graph_def of the model to be scaled
        alpha: float, the scaling factor
        scales_per_op: bool, each op will have an individual scale or
                       ops with the same input will share a scale
        op_types:
    """

    def __init__(self, graph_def, alpha, scales_per_op, op_types):
        """Initialization."""
        self.graph_def = graph_def
        self.alpha = alpha
        self.scales_per_op = scales_per_op
        self.op_types = op_types

        self.graph_info = None
        self.mul_list = []
        self.sq_weight_scale_dict = {}

    def _parse_weight_dict(self, max_vals_per_channel, sq_weight_tensor_dict):
        """Parse weight related dictionaries to two required dictionaries.

        Args:
            max_vals_per_channel (dict): A dictionary containing the maximum values per channel.
            sq_weight_tensor_dict (dict): A dictionary containing tensor of weights.

        Returns:
            sq_weight_tensors: A dictionary whose structure is like {input_node_name: weight_tensor}}.
            sq_weights_node_names: A dictionary whose structure is like {input_node_name: weight_node_name}}.
        """
        sq_weight_tensors = {}
        sq_weight_node_names = {}
        for input_node_name in max_vals_per_channel:
            curr_weight_tensors = []
            curr_weights_node_names = []
            next_node_names = self.graph_info[input_node_name].outputs
            for node_name in next_node_names:
                curr_node = self.graph_info[node_name].node
                if curr_node.op not in self.op_types:
                    continue
                if len(curr_node.input) >= 2:
                    weight_name = curr_node.input[1]
                    weight_tensor = sq_weight_tensor_dict[weight_name]
                    curr_weight_tensors.append(weight_tensor)
                    curr_weights_node_names.append(weight_name)
            sq_weight_tensors[input_node_name] = curr_weight_tensors
            sq_weight_node_names[input_node_name] = curr_weights_node_names
        return sq_weight_tensors, sq_weight_node_names

    def transform(self, max_vals_per_channel, sq_weight_tensor_dict, sq_target_node_names):
        """Apply scaling to weights and activations based on the maximum values per channel.

        Args:
            max_vals_per_channel (dict): A dictionary containing the maximum values per channel for each input node.
            sq_weight_tensor_dict (dict): A dictionary whose structure is like {input_node_name: weight_tensor}.
            sq_target_node_names (dict): A dictionary whose structure is like {weight_node_name: target_node_name}.
        """
        self.g_analyzer = GraphAnalyzer()
        self.g_analyzer.graph = self.graph_def
        self.graph_info = self.g_analyzer.parse_graph()
        sq_weight_tensors, sq_weight_node_names = self._parse_weight_dict(max_vals_per_channel, sq_weight_tensor_dict)
        logger.info("Start scaling on model graph for Smooth Quantization.")
        if self.scales_per_op:
            # 1. obtain the smooth scale per op
            # 2. adjust weight
            # 3. adjust activation
            for _, input_node_name in enumerate(max_vals_per_channel):
                activation_max_per_in_channel = max_vals_per_channel[input_node_name]
                W_lst = sq_weight_tensors[input_node_name]  # VQK weight value
                # Use the const nodes before to get weight values, VQK ReadVariable
                W_node_name_lst = sq_weight_node_names[input_node_name]
                # Get the concrete weight node as the output of Mul insertion, QKV ReadVariable
                for w_i, W in enumerate(W_lst):
                    if len(W.shape) == 4:
                        # https://www.tensorflow.org/api_docs/python/tf/nn/conv2d
                        # weight: [filter_height, filter_width, in_channels, out_channels]
                        # activation: NHWC, also batch_shape + [in_height, in_width, in_channels]
                        tensor = np.abs(np.transpose(W, [0, 1, 3, 2]))
                        # reduce weight max to (in_channel, ), aligned with activation max
                        W_max_per_in_channel = np.max(np.reshape(tensor, (-1, tensor.shape[-1])), axis=0)
                    elif len(W.shape) == 2:  # matmul
                        # reduce weight max to (in_channel, ), aligned with activation max
                        tensor = np.abs(W)
                        W_max_per_in_channel = np.max(tensor, axis=1)
                    else:  # pragma: no cover
                        assert False, "not supported"
                    cur_weight_node_name = W_node_name_lst[w_i]
                    try:
                        scale = np.power(activation_max_per_in_channel, self.alpha) / np.power(
                            W_max_per_in_channel, (1 - self.alpha)
                        )
                    except ValueError as e:  # pragma: no cover
                        logger.info(e)
                        logger.info("Skip smoothing the node: {}".format(cur_weight_node_name))
                        continue
                    # clip the scales that are too small
                    scale = np.clip(scale, a_min=1e-5, a_max=1e8)
                    # skip smoothing the op where scale has elements that less than 1
                    # if np.any(scale < 1):
                    #     logger.info("skip smooth quant: {}".format(input_node_name))
                    #     continue
                    self.sq_weight_scale_dict[cur_weight_node_name] = scale
                    self._adjust_activation(1 / scale, input_node_name, sq_target_node_names[cur_weight_node_name], w_i)
        else:
            pass
        sq_graph_def = self.g_analyzer.dump_graph()
        sq_graph_def.library.CopyFrom(self.graph_def.library)
        return sq_graph_def, self.sq_weight_scale_dict, self.mul_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse Decomposed BatchNorm Graph Rewriter."""

import collections
import math
import re

import numpy as np
from tensorflow.compat.v1 import graph_util
from tensorflow.core.framework import attr_value_pb2, graph_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util
from tensorflow.python.platform import flags as flags_lib
from tensorflow.python.platform import tf_logging
from tensorflow.python.tools import strip_unused_lib

from neural_compressor.tensorflow.utils import dump_elapsed_time


class FuseDecomposedBNOptimizer:
    """Fuse decomposed small ops to BatchNormalization."""

    def __init__(self, input_graph_def):
        """Initialization."""
        self.input_graph_def = input_graph_def

    @dump_elapsed_time("Pass FuseDecomposedBNOptimizer")
    def do_transformation(self):
        """Fuse individual ops in batch normalization to FusedBatchNorm.

        In some models, the batch normalizatin is performed via a group of individual
        ops instead of using single FusedBatchNorm op. This function identifies a
        pattern of batch normalization subgraph which is made of multiple ops and
        transforms the graph by replacing those individual ops with FusedBatchNorm op.
        This will provide the opportunity to further fold the FusedBatchNorm with
        convolution ops to reduce the computation steps during inference.
        This function currently recognizes batch normalization patterns described
        below, this could be extended if newer patterns are seen. Also, the fusion
        is only attempted if the input graph is in NHWC format or has no format set.
        Computation function:
            (X * multiplier) + (Beta - Mean * multiplier)
            where multiplier = rsqrt (Variance + Epsilon) * Gamma
                            OR = rsqrt (Variance + Epsilon) when Gamma is 1
        Subgraph:
        {"Add"
            {{"Mul"  // mul_0
                {{"*"},  // input to apply batchnorm
                {"Mul"  // mul_1, same op is used inside the Sub block
                    {{"Rsqrt"
                        {"Add"
                            {{"Const" | "Reshape(Const)"},  // Variance
                            {"Const"}  // Epsilon
                            }
                        }
                        },  // end - Rsqrt
                        {"Const" | "Reshape(Const)"}  // Gamma
                    }
                    }  // end - mul_1
                }
            },  // end - mul_0
            {"Sub"
                {{"Const" | "Reshape(Const)"},  // Beta
                {"Mul"  // mul_3
                    {{"Const" | "Reshape(Const)"},  // Mean
                    {"Mul"  // same mul_1 op as in previous block
                        {{"Rsqrt"
                            {"Add"
                                {{"Const" | "Reshape(Const)"},  // Variance
                                {"Const"}  // Epsilon
                                }
                            }
                        },  // end - Rsqrt
                        {"Const" | "Reshape(Const)"}  // Gamma
                        }
                        }  // end - mul_1
                    }
                    }  // end - mul_3
                }
                }  // end - Sub
            }
        }  // end - Add
        Subgraph pattern when gamma value is 1 and the gamma scaling Mul is skipped
        {"Add"
            {{"Mul"  // mul_0
                {{"*"},  // input to apply batchnorma
                {"Rsqrt"  // same Rsqrt op used in Sub block
                    {"Add"
                        {{"Const" | "Reshape(Const)"},  // Variance
                        {"Const"}  // Epsilon
                        }
                    }
                    }  // end - Rsqrt
                }
                },  // end - mul_0
                {"Sub"
                {{"Const" | "Reshape(Const)"},  // Beta
                {"Mul"  // mul_1
                    {{"Const" | "Reshape(Const)"},  // Mean
                    {"Rsqrt"  // same Rsqrt op as in previous mul_0 block
                        {"Add"
                            {{"Const" | "Reshape(Const)"},  // Variance
                            {"Const"}  // Epsilon
                            }
                        }
                        }  // end - Rsqrt
                    }
                }  // end - mul_1
                }
                }  // end - Sub
            }
        }  // end - Add
        Args:
            input_graph_def: A GraphDef containing a model.

        Returns:
            Modified graph with individual ops that made up of batch normalization
            fused to FusedBatchNorm.

        Raises:
            ValueError: If the graph is badly formed with duplicate node names.
        """
        input_node_map = {}
        for node in self.input_graph_def.node:
            if node.name not in input_node_map:
                input_node_map[node.name] = node
            else:
                raise ValueError("Duplicate node names detected for ", node.name)

        # Check format and only proceed if graph is in NHWC or has no format set.
        data_format = None
        for node in self.input_graph_def.node:
            if "data_format" in node.attr.keys():
                data_format = node.attr["data_format"]
            if data_format is not None and data_format.s != b"NHWC":
                tf_logging.warn("%s in %s format, not candidate for batchnorm fusion." % (node.name, data_format.s))
                return self.input_graph_def
            else:
                continue

        nodes_to_skip = {}
        new_ops = []
        for node in self.input_graph_def.node:
            if node.op != "Add":
                continue

            # Add (Mul, Sub) or Add (Sub, Mul)
            input0_op = node_from_map(input_node_map, node.input[0])
            input1_op = node_from_map(input_node_map, node.input[1])

            if input0_op.op == "Mul" and input1_op.op == "Sub":
                data_scale_mul_op = input0_op
                bias_mean_sub_op = input1_op
            elif input0_op.op == "Sub" and input1_op.op == "Mul":
                bias_mean_sub_op = input0_op
                data_scale_mul_op = input1_op
            else:
                continue

            # Mul (input, Mul)
            input_data_op = node_from_map(input_node_map, data_scale_mul_op.input[0])
            # Workaround for model ava-person-vehicle-detection-stage2-2_0_0
            # FusedBatchNorm requires a 4D Tensor for input data,
            # but the MatMul before FusedBatchNorm only support 2D output.
            # Don't fuse the small ops to FusedBatchNorm when the upstream has MatMul.
            if input_data_op.op == "MatMul":
                continue

            # Workaround for DIEN_Deep-Interest-Evolution-Network
            if input_data_op.op == "ConcatV2" and input_data_op.name == "concat_8":
                continue

            if input_data_op.input:
                ancestor_input_data_op = node_from_map(input_node_map, input_data_op.input[0])
                if ancestor_input_data_op.op == "MatMul":
                    continue

            scale_op = node_from_map(input_node_map, data_scale_mul_op.input[1])

            if scale_op.op == "Rsqrt":
                gamma_op = None
                gamma_reshape_op = None
                rsqrt_op = scale_op
            elif scale_op.op == "Mul":
                # Mul (Rsqrt, Constant_gamma)
                rsqrt_op = node_from_map(input_node_map, scale_op.input[0])
                gamma_op, gamma_reshape_op = bypass_reshape(input_node_map, scale_op.input[1])
                if rsqrt_op.op != "Rsqrt":
                    continue
                if gamma_op.op != "Const" or get_const_dim_count(gamma_op) != 1:
                    continue
            else:
                continue

            # Sub (Constant_beta, Mul)
            beta_op, beta_reshape_op = bypass_reshape(input_node_map, bias_mean_sub_op.input[0])
            mean_scale_mul_op = node_from_map(input_node_map, bias_mean_sub_op.input[1])
            if mean_scale_mul_op.op != "Mul":
                continue
            if beta_op.op != "Const" or get_const_dim_count(beta_op) != 1:
                continue

            # Common scale applies to both input and running mean
            if scale_op != node_from_map(input_node_map, mean_scale_mul_op.input[1]):
                continue

            mean_op, mean_reshape_op = bypass_reshape(input_node_map, mean_scale_mul_op.input[0])
            if mean_op.op != "Const" or get_const_dim_count(mean_op) != 1:
                continue

            # Add (Constant_variance, Constant_epsilon)
            variance_epsilon_add_op = node_from_map(input_node_map, rsqrt_op.input[0])
            if variance_epsilon_add_op.op != "Add":
                continue

            variance_op, variance_reshape_op = bypass_reshape(input_node_map, variance_epsilon_add_op.input[0])
            epsilon_op = node_from_map(input_node_map, variance_epsilon_add_op.input[1])
            if epsilon_op.op != "Const" or get_const_dim_count(epsilon_op) != 0:
                continue
            if variance_op.op != "Const" or get_const_dim_count(variance_op) != 1:
                continue

            epsilon = values_from_const(epsilon_op)

            nodes_to_skip[node.name] = True
            nodes_to_skip[data_scale_mul_op.name] = True
            nodes_to_skip[bias_mean_sub_op.name] = True
            nodes_to_skip[mean_scale_mul_op.name] = True
            nodes_to_skip[scale_op.name] = True
            if scale_op.op != "Rsqrt":
                nodes_to_skip[rsqrt_op.name] = True
            nodes_to_skip[variance_epsilon_add_op.name] = True
            if gamma_reshape_op is not None:
                nodes_to_skip[gamma_reshape_op.name] = True
            if beta_reshape_op is not None:
                nodes_to_skip[beta_reshape_op.name] = True
            if mean_reshape_op is not None:
                nodes_to_skip[mean_reshape_op.name] = True
            if variance_reshape_op is not None:
                nodes_to_skip[variance_reshape_op.name] = True

            if gamma_op is None:
                gamma_op = node_def_pb2.NodeDef()
                gamma_op.op = "Const"
                # Assign name with same root of Rsqrt op's name plus "gamma"
                m = re.search(r"(.*)/(.*)", scale_op.name)
                if m:
                    gamma_op.name = m.group(1) + "/gamma"
                else:
                    gamma_op.name = scale_op.name + "/gamma"
                gamma_op.attr["dtype"].CopyFrom(beta_op.attr["dtype"])
                beta_value = values_from_const(beta_op)
                gamma_op.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(
                            1, beta_value.dtype.type, beta_value.shape, allow_broadcast=True
                        )
                    )
                )
                new_ops.append(gamma_op)

            new_fused_batchnorm_op = node_def_pb2.NodeDef()
            new_fused_batchnorm_op.op = "FusedBatchNorm"
            new_fused_batchnorm_op.name = node.name
            new_fused_batchnorm_op.attr["T"].CopyFrom(node.attr["T"])
            new_fused_batchnorm_op.attr["is_training"].CopyFrom(attr_value_pb2.AttrValue(b=False))
            new_fused_batchnorm_op.attr["epsilon"].CopyFrom(attr_value_pb2.AttrValue(f=epsilon.tolist()))
            if data_format is not None:
                new_fused_batchnorm_op.attr["data_format"].CopyFrom(data_format)
            new_fused_batchnorm_op.input.extend(
                [input_data_op.name, gamma_op.name, beta_op.name, mean_op.name, variance_op.name]
            )

            new_ops.append(new_fused_batchnorm_op)

        result_graph_def = graph_pb2.GraphDef()
        for node in self.input_graph_def.node:
            if node.name in nodes_to_skip:
                continue
            new_node = node_def_pb2.NodeDef()
            new_node.CopyFrom(node)
            retained_input = []
            for input_node in new_node.input:
                if not input_node.startswith("^") or input_node[1:] not in nodes_to_skip:
                    retained_input.append(input_node)
            new_node.input[:] = retained_input
            result_graph_def.node.append(new_node)

        result_graph_def.node.extend(new_ops)
        result_graph_def.versions.CopyFrom(self.input_graph_def.versions)
        return result_graph_def


def node_name_from_input(node_name):
    """Strips off ports and other decorations to get the underlying node name."""
    if node_name.startswith("^"):
        node_name = node_name[1:]
    m = re.search(r"(.*):\d+$", node_name)
    if m:
        node_name = m.group(1)
    return node_name


def node_from_map(node_map, name):
    """Pulls a node def from a dictionary for a given name.

    Args:
        node_map: Dictionary containing an entry indexed by name for every node.
        name: Identifies the node we want to find.

    Returns:
        NodeDef of the node with the given name.

    Raises:
        ValueError: If the node isn't present in the dictionary.
    """
    stripped_name = node_name_from_input(name)
    if stripped_name not in node_map:
        raise ValueError("No node named '%s' found in map." % name)
    return node_map[stripped_name]


def values_from_const(node_def):
    """Extracts the values from a const NodeDef as a numpy ndarray.

    Args:
        node_def: Const NodeDef that has the values we want to access.

    Returns:
        Numpy ndarray containing the values.

    Raises:
        ValueError: If the node isn't a Const.
    """
    if node_def.op != "Const":
        raise ValueError("Can not extract constant value from a node that is not Const. Got:\n" f"{node_def}")
    input_tensor = node_def.attr["value"].tensor
    tensor_value = tensor_util.MakeNdarray(input_tensor)
    return tensor_value


def valid_reshape_inputs(reshape_in0_ndef, reshape_in1_ndef):
    """Check if the inputs of the Reshape are valid."""
    if reshape_in0_ndef.op != "Const" or reshape_in1_ndef.op != "Const" or get_const_dim_count(reshape_in0_ndef) != 1:
        return False
    input0_vec_size = values_from_const(reshape_in0_ndef).shape[0]
    const_value = values_from_const(reshape_in1_ndef)
    shape_ndims = const_value.ndim
    if shape_ndims != 1:
        raise ValueError("Num of dims of the shape must be 1, got {}.".format(shape_ndims))
    for value in const_value.tolist()[:-1]:
        if value != 1:
            return False
    if const_value.tolist()[-1] != input0_vec_size:
        return False
    return True


def bypass_reshape(input_node_map, input_name):
    """Get Reshape input nodes."""
    reshape_ndef = None
    maybe_reshape_ndef = node_from_map(input_node_map, input_name)
    input_ndef = maybe_reshape_ndef
    if maybe_reshape_ndef.op == "Reshape":
        reshpae_input0_ndef = node_from_map(input_node_map, maybe_reshape_ndef.input[0])
        reshpae_input1_ndef = node_from_map(input_node_map, maybe_reshape_ndef.input[1])
        if (
            reshpae_input0_ndef.op == "Const"
            and reshpae_input1_ndef.op == "Const"
            and valid_reshape_inputs(reshpae_input0_ndef, reshpae_input1_ndef)
        ):
            input_ndef = reshpae_input0_ndef
            reshape_ndef = maybe_reshape_ndef
    return input_ndef, reshape_ndef


def get_const_dim_count(node_def):
    """Get the number of dimensions for a Const node.

    Args:
        node_def: Const NodeDef.

    Returns:
        Number of dimensions for the Const node.
    """
    const_value = values_from_const(node_def)
    return const_value.ndim
